import numpy as np
import pandas as pd
import random
from kero.DataHandler.Generic import *


def dataframe_to_list(dataframe):
    # convert a data frame to list
    # WARNING: each element in the output list is an entire column
    #
    # -- example usage --
    #
    # First, save a dataframe by the file name "ggtest.csv"
    # see example in class RandomDataFrame at RandomDataFrame.py. Then
    #
    # rdf = RDF.RandomDataFrame()
    # overviewer = dt.data_overview()
    # overviewer.original_df = pd.read_csv(r"ggtest.csv")
    # dfnow = overviewer.original_df[:]
    # print(overviewer.original_df)
    #
    # dflist = dt.dataframe_to_list(dfnow)
    # print(type(dflist))
    # for x in dflist:
    #     print(x)

    df = dataframe[:]
    listform = []
    # print(df)
    column_name_list = df.columns
    # print(column_name_list)
    for colname in column_name_list:
        # print(colname)
        temp=[]
        col=df[colname]
        for x in col:
            temp.append(x)
        listform.append(temp)
    column_name_list = df.columns
    return listform, column_name_list


def column_defect_index(inputlist):
    # inputlist: list
    # returns list of indices where inputlist is Nan, np.nan
    # !! do add other forms of defects
    defect_indices=[]
    for i in range(len(inputlist)):
        x=inputlist[i]
        if isinstance(x,int) or isinstance(x,float):
            if np.isnan(float(x)):
                defect_indices.append(i)
                continue
        if isinstance(x,str):
            if x=='':
                defect_indices.append(i)
                continue
    return defect_indices


def table_defect_index(dataframe_in_list):
    # dataframe_in_list : list of list0. It is dataframe transformed into a list
    # each list0 is a list, corresponding to an entire column in a dataframe; transpose accordingly
    # return : list of indices of rows where there is Nan etc
    all_indices = []
    for col in dataframe_in_list:
        all_indices = all_indices+column_defect_index(col)
    return list(sorted(set(all_indices), key=all_indices.index))


def drop_defective_table(dataframe):
    # dataframe : panda dataframe
    # return (c_data,cr_data) where
    # c_data is subset of dataframe without defective entries
    # cr_data is subset of dataframe with defective entries
    df = dataframe[:]
    # print(df)
    dflist, column_list_name = dataframe_to_list(dataframe)  # list of x, x is a list, which is an entire column
    defect_ind = table_defect_index(dflist)
    defect_ind = np.sort(defect_ind)
    # print("defect_ind : ",defect_ind )
    dflist = transpose_list(dflist)
    # n_col=len(dflist[0])
    n_row = len(dflist)
    clean_list = []
    crippled_list = []
    if len(defect_ind) == 0:
        return df, None
    else:
        min_ind = defect_ind[0]
        for i in range(n_row):
            if i == min_ind:
                crippled_list.append(dflist[i])
                if len(defect_ind) == 1:
                    pass
                else:
                    defect_ind = defect_ind[1:]
                    min_ind = defect_ind[0]
            else:
                clean_list.append(dflist[i])

        # print(clean_list)
        # print(crippled_list)
        col_names = dataframe.columns
        crippled_df = pd.DataFrame(crippled_list, columns=col_names)
        clean_df = pd.DataFrame(clean_list, columns=col_names)

    return clean_df, crippled_df


def create_list_of_empty_list(n=2):
    out = []
    for i in range(n):
        out.append([])
    return out


def data_sieve(dataframe):
    df = dataframe[:]
    cleanD = clean_data()
    crippD = crippled_data()
    origD = original_data()

    origD.original_df = df
    clean_df, crippled_df = drop_defective_table(df)
    cleanD.clean_df = clean_df
    crippD.crippled_df = crippled_df
    return cleanD, crippD, origD


def get_list_clean(input_list):
    # output the same list, without defective entries e.g. NaN etc
    defective_ind = list(np.sort(column_defect_index(input_list)))
    out = []
    for i in range(len(input_list)):
        if len(defective_ind) > 0:
            if i == defective_ind[0]:
                defective_ind = defective_ind[1:]
            else:
                out.append(input_list[i])
        else:
            out.append(input_list[i])
    return out


# Suggested naming convention.
# data object contains dataframes and their transformations
# name the variable storing the object with suffix D
# for example, cleanD = clean_data()
class original_data:
    def __init__(self):
        self.original_df = None
        self.repaired_df = None
        # list version
        self.original_list = None
        self.repaired_list = None
        return

    def initialize_dataframe_repair(self):
        self.repaired_df = self.original_df.copy()
        return

    def repair_table(self, repair_command_settings=None):
        # -- repair_command_settings --
        # dictionary={column_name1: mode1, ...}
        if repair_command_settings is None:
            print("repair_table(): there is no instruction for table reparation.")
            return
        column_keys = repair_command_settings.keys()
        for column_key in column_keys:
            # print("key = ",col_key, "\nmode = ", repair_command_settings[column_key])
            mode = repair_command_settings[column_key]
            self.repair_single_column(column_key, mode=mode)
            # pass
        return

    def repair_single_column(self, column_key, mode=None):
        # --column_key--
        # string: name of column
        # --mode--
        # see the choices below

        if self.repaired_df is None:
            print("repair_single_column(): self.repaired_df is None."
                  "\nReminder to perform initialize_dataframe_repair()")
            return
        this_column = list(self.repaired_df[column_key])
        defective_index = list(np.sort(column_defect_index(this_column)))
        clean_column = get_list_clean(this_column)
        # Choices :
        # 1. mean
        # 2. mean_floor
        # 3. mean_ceil
        # 4. max_occuring
        # 5. min_occuring
        # 6. mid_occuring
        if mode == "mean":
            mean_padding_value = np.mean(clean_column)
            for i in range(len(defective_index)):
                self.repaired_df.at[defective_index[i], column_key] = mean_padding_value
        elif mode == "mean_floor":
            mean_padding_value = np.floor(np.mean(clean_column))
            for i in range(len(defective_index)):
                self.repaired_df.at[defective_index[i], column_key] = mean_padding_value
        elif mode == "mean_ceil":
            mean_padding_value = np.ceil(np.mean(clean_column))
            for i in range(len(defective_index)):
                self.repaired_df.at[defective_index[i], column_key] = mean_padding_value
        elif mode == "max_occuring":
            lstat = ListStats(mylist=clean_column)
            lstat.prepare_list()
            max_occuring = lstat.max_occuring
            secure_random = random.SystemRandom()
            padding_value = secure_random.choice(max_occuring)
            for i in range(len(defective_index)):
                self.repaired_df.at[defective_index[i], column_key] = padding_value
            pass
        elif mode == "min_occuring":
            lstat = ListStats(mylist=clean_column)
            lstat.prepare_list()
            min_occuring = lstat.min_occuring
            secure_random = random.SystemRandom()
            padding_value = secure_random.choice(min_occuring)
            for i in range(len(defective_index)):
                self.repaired_df.at[defective_index[i], column_key] = padding_value
            pass
        elif mode == "mid_occuring":
            lstat = ListStats(mylist=clean_column)
            lstat.prepare_list()
            mid_occuring = lstat.mid_occuring
            secure_random = random.SystemRandom()
            for i in range(len(defective_index)):
                padding_value = secure_random.choice(mid_occuring)
                self.repaired_df.at[defective_index[i], column_key] = padding_value
            pass
        else:
            pass

        # print("this column = ", this_column)
        # print("clean column = ", clean_column)
        return

class clean_data:
    def __init__(self):
        self.clean_df=None # data frame without Nan etc
        self.clean_df_conj=None
        self.colname_set=None # set of strings containing column names

        # -- list versions of the df --
        self.clean_list=None
        self.clean_list_conj=None
        self.colname_set_conj=None

        self.column_module_set={}
        return

    def get_list_from_df(self):
        self.clean_list=dataframe_to_list(self.clean_df) # each entry is a column
        self.clean_list=transpose_list(self.clean_list) # each entry a row
        return

    def build_column_module_set(self, conj_command_set, conj_command_setting_set=None):
        # conj_command_set: set of strings
        # - [com1,...,comN] , where N is the number of columns in the dataframe
        # - for index j, perform comj on column j where com j are:
        # - 1. "discrete_to_bool"
        # - 2. "cont_to_scale"
        # - 3. "discrete_to_int"
        # - 4. ""
        # see set_conj functions in column_module class objects
        # conj_command_setting_set: extra settings
        # - 1
        # - 2
        # - 3 rule : a dictionary specifying for each element,
        #   which integer to transform into
        #   for example {"a":1,"b":22,"c":7}
        # - 4

        self.colname_set=self.clean_df.columns
        # print("build_column_module_set(). self.colname_set =  ", self.colname_set)
        current_index=0
        # for colname in self.colname_set:
        for i in range(0,len(self.colname_set)):
            colname = self.colname_set[current_index]
            self.column_module_set[colname] = column_module()
            current_command = conj_command_set[colname]
            self.column_module_set[colname].get_list_from_df(self.clean_df, current_index)
            if current_command == "discrete_to_bool":
                # recall: this is one-hot
                if conj_command_setting_set is None:
                    self.column_module_set[colname].set_conj_discrete_to_bool()
                else:
                    drop_one_column = conj_command_setting_set[colname]
                    # print("build_column_module_set()"
                    #       "one hot drop column setting =", drop_one_column)
                    self.column_module_set[colname].set_conj_discrete_to_bool(drop_one_column=drop_one_column)
            elif current_command == "cont_to_scale":
                if conj_command_setting_set is None:
                    self.column_module_set[colname].set_conj_cont_to_scale()
                else:
                    settings = conj_command_setting_set[colname]
                    # typical settings = {"scale": [-1, 1], "mode": "uniform"}, a dictionary
                    scale = None
                    mode = None
                    original_scale = None
                    if settings is not None:
                        scale = settings["scale"]
                        mode = settings["mode"]
                        original_scale = settings["original_scale"]

                    # print("build_column_module_set(): scale / mode =", scale, "/", mode)
                    self.column_module_set[colname].set_conj_cont_to_scale(scale=scale,original_scale=original_scale, mode=mode)
                    pass
            elif current_command == "discrete_to_int":
                if conj_command_setting_set is None:
                    self.column_module_set[colname].set_conj_discrete_to_int()
                else:
                    newrule=conj_command_setting_set[colname]
                    try:
                        self.column_module_set[colname].set_conj_discrete_to_int(rule=newrule)
                    except:
                        self.column_module_set[colname].set_conj_discrete_to_int()
                        print("build_column_module_set: do check out if there is problem"
                              "in the discrete_to_int section")
            else:
                pass
            current_index=current_index+1
        return

    def build_conj_dataframe(self, conj_command_set, conj_command_setting_set=None):
        # conj dataframe is dataframe whose each column is transformed according to conj_command_set
        self.build_column_module_set(conj_command_set, conj_command_setting_set=conj_command_setting_set)
        self.clean_list_conj = []
        self.colname_set_conj = []
        current_index = 0
        link = ""
        # for col_mod_key in self.column_module_set: # this is disordered
        for kk in range(len(self.colname_set)):
            col_mod_key = self.column_module_set[self.colname_set[kk]].col_name
            # print("build_conj_dataframe(). column:", col_mod_key)
            # print(self.column_module_set[col_mod_key].col_list)
            # print("build_conj_dataframe(). conjugate:", col_mod_key)
            # print(self.column_module_set[col_mod_key].col_list_conj)     
            if conj_command_set[col_mod_key] == "discrete_to_bool":
                # pass
                temp=self.column_module_set[col_mod_key].col_list_conj
                # print(temp.keys())
                # print(temp)
                for sub_col_key in temp:
                    self.colname_set_conj.append(link.join((col_mod_key,"_",sub_col_key)))
                    self.clean_list_conj.append(temp[sub_col_key])
            elif conj_command_set[col_mod_key]=="cont_to_scale":
                # pass
                temp = self.column_module_set[col_mod_key].col_list_conj
                # print(temp)
                self.clean_list_conj.append(temp[:])
                self.colname_set_conj.append(col_mod_key)
            elif conj_command_set[col_mod_key]=="discrete_to_int":
                temp = self.column_module_set[col_mod_key].col_list_conj
                self.clean_list_conj.append(temp[:])
                self.colname_set_conj.append(col_mod_key)
                # pass
            else:
                # No transformation is effected
                # Initial column is retained
                # pass
                self.colname_set_conj.append(col_mod_key)
                temp=self.column_module_set[col_mod_key].col_list[:]
                # print(temp)
                self.clean_list_conj.append(temp)
            current_index=current_index+1
        # print("build_conj_dataframe(). MARKER")
        self.clean_list_conj = transpose_list(self.clean_list_conj)
        self.clean_df_conj = pd.DataFrame(self.clean_list_conj, columns=self.colname_set_conj)
        # try:
        #     self.clean_df_conj=pd.DataFrame(self.clean_list_conj,columns=self.colname_set_conj)
        # except Exception as e:
        #     print(e)
        #     print('length=',len(self.colname_set_conj)," : ",self.colname_set_conj)
        #     count=0
        #     for x in self.clean_list_conj:
        #         print(x)
        #         count=count+1
        #     print("length=",count)
        return


class crippled_data:
    def __init__(self):
        self.crippled_df = None  # data frame with defects

        # -- list versions of the df --
        self.crippled_list = None
        return

# column handler
# A more detailed handler, each handling only one column

class column_module:
    # assume clean data ONLY, i.e. no Nan etc
    def __init__(self):
        self.col_name=None
        self.col_list=None # a single list

        # transformed list
        # Each conj functions, i.e. a function that generates the conj
        # will describe the prototype. Check them out.
        self.col_list_conj = None  # IMPT: a dictionary
        return


    def get_list(self, col_list, col_name):
        # alternatively, use get_list_from_df(), more convenient given a dataframe
        self.col_list=col_list
        self.col_name = col_name
        return


    def get_list_from_df(self,df,index):
        # df : panda dataframe
        # index : int, the column (0-based index) which this object extracts from df for storing
        df_columns=list(df.columns)
        self.get_list(list(df[df_columns[index]]), df_columns[index])
        return
    # types of conj to transform into


    def set_conj_discrete_to_bool(self,drop_one_column=False):
        temp=self.col_list[:]
        temp=[str(x) for x in temp]
        self.col_list_conj = conj_from_discrete_to_bool(temp, drop_one_column=drop_one_column)
        return

    def set_conj_cont_to_scale(self, scale=None, original_scale=None, mode="uniform"):
        temp = self.col_list[:]
        if scale is None:
            scale = [0, 1]
        self.col_list_conj = conj_from_cont_to_scaled(temp, scale=scale, 
            original_scale=original_scale, mode=mode)
        return

    def set_conj_discrete_to_int(self, rule=None):
        temp = self.col_list[:]
        self.col_list_conj = conj_from_discrete_to_int(temp, rule=rule)
        return


def conj_from_discrete_to_bool(col, drop_one_column=False):
    # This is the so-called ONE-HOT!
    # col: a list of string IMPT: make sure to make them string
    # drop_one_column=: drop the last column, prevent dummy trap variables from affecting results
    #
    # -- prototype --
    # For column which takes in discrete values, convert it to columns of Bool
    # for example:
    # 1. string. e.g. nationality, "id","my","sg". The conjugate will be 3 columns
    #    respectively; is_id, is_my, is_sg
    # 2, discrete numbers. e.g. number of children, "0","1","2+". Conjugate will be is_0, is_1, is_2p
    #
    # return {is_0 : [1,0,0,1,0], is_1 : [0,1,1,0,0], is_2p : [0,0,0,0,1]}

    # recall: data is clean
    unique_el=list(sorted(set(col)))
    dict_output={}
    for ue in unique_el:
        dict_output[str(ue)]=[0 for x in range(len(col))]
    count=0
    for x in col:
        dict_output[x][count]=1
        count=count+1
    if drop_one_column:
        drop_key=unique_el[-1]
        dict_output.pop(drop_key)
    return dict_output


def conj_from_cont_to_scaled(col, scale=None, original_scale=None, mode="uniform"):
    # col: a list of float, continuous values
    # scale: [min_value, max_value]
    # mode: string, to set the mode
    #
    # -- prototype --
    # for example, col=[1,2,10], scaled with scale=[0,1] in uniform mode
    #
    # return [0.1, 0.2, 1.0]
    if scale is None:
        scale = [0,1]
    out = []
    if mode == "uniform":
        min_val = scale[0]
        max_val = scale[1]
        if original_scale is None:
            min_raw = np.min(col)
            max_raw = np.max(col)
        else:
            min_raw = original_scale[0]
            max_raw = original_scale[1]
        for x in col:
            temp = (x-min_raw)/(max_raw-min_raw)
            temp = min_val + temp * (max_val-min_val)
            out.append(temp)
    else:
        out = col
    return out


def conj_from_discrete_to_int(col, rule=None):
    # col: a list of string IMPORTANT: make sure to make them string
    #
    # -- prototype --
    # label each item with an integer. Same item is labelled with same integer
    # for example, if col=["a","a","b","c","b"]
    # return [0,0,1,2,1]

    # rule example {"a":100,"b":200,"c":7}
    # return [100,100,200,7,200]

    # recall: data is clean

    out = []
    if rule is None:
        uel = list(sorted(set(col), key=col.index))
        for x in col:
            temp = uel.index(x)
            out.append(temp)
    else:
        for x in col:
            temp = rule[x]
            out.append(temp)
    return out


