import numpy as np
import pandas as pd
import kero.DataHandler.RandomDataFrame as RDF
import kero.DataHandler.DataTransform as dt
import kero.DataHandler.DataVisual as dv
import kero.DataHandler.Generic as gen
import matplotlib.pyplot as plt


# Check that initiate_random_table() works
def check_initiate_random_table(csv_name="check_table_defect_index.csv",
                                rate=0.14, output_label=None, with_unique_ID=None):
    rdf = RDF.RandomDataFrame()
    col1 = {"column_name": "first", "items": [1, 2, 3]}
    itemlist = list(np.linspace(10, 20, 48))
    col2 = {"column_name": "second", "items": itemlist}
    col3 = {"column_name": "third", "items": ["gg", "not"]}
    col4 = {"column_name": "fourth", "items": ["my", "sg", "id", "jp", "us", "bf"]}

    if output_label is not None:
        if output_label=="classification":
            col_out={"column_name": "result", "items": ["classA","classB","classC"]}
            rdf.initiate_random_table(20, col1, col2, col3, col4,col_out, panda=True)
    else:
        rdf.initiate_random_table(20, col1, col2, col3, col4, panda=True,with_unique_ID=with_unique_ID)
    rdf.crepify_table(rdf.clean_df, rate=rate)
    try:
        rdf.crepified_df.to_csv(csv_name, index=False)
    except:
        print("check_initiate_random_table. Error.")
    return rdf


# Check table_defect_index() is correct:
# recall: it returns all indices of rows with defects
def check_table_defect_index():
    rdf = check_initiate_random_table()
    df = pd.read_csv(r"check_table_defect_index.csv")
    listform = dt.dataframe_to_list(df)
    print(df)
    all_index = dt.table_defect_index(listform)
    print("Defective indices:\n", all_index)
    return

# Check drop_defective_table() is correct:
def check_drop_defective_table():
    rdf = check_initiate_random_table()
    df = pd.read_csv(r"check_table_defect_index.csv")
    dt.drop_defective_table(df)
    clean_data, crippled_data = dt.drop_defective_table(df)

    print(df)
    print("\nclean df:\n")
    print(clean_data)
    print("\ncrippled df:\n")
    print(crippled_data)
    return

def check_data_sieve():
    rdf = check_initiate_random_table(csv_name="check_data_sieve.csv")
    df = pd.read_csv(r"check_data_sieve.csv")
    cleanD, crippD,_=dt.data_sieve(df)
    print(df)
    print("\nclean df:\n")
    print(cleanD.clean_df)
    print("\ncrippled df:\n")
    print(crippD.crippled_df)
    return

def check_prepare_for_visual():
    rdf = check_initiate_random_table(csv_name="check_prepare_for_visual.csv")
    df = pd.read_csv(r"check_prepare_for_visual.csv")
    vlab=dv.visual_lab()
    vlab.prepare_for_visual(df)
    print(vlab.df)
    print("\nclean df:\n")
    print(vlab.cleanD.clean_df)
    print("\ncrippled df:\n")
    print(vlab.crippledD.crippled_df)
    return

def check_inspect_original_data():
    rdf = check_initiate_random_table(csv_name="check_inspect_original_data.csv")
    df = pd.read_csv(r"check_inspect_original_data.csv")
    vlab = dv.visual_lab()
    vlab.prepare_for_visual(df)
    vlab.inspect_original_data(txt_report_filename="check_inspect_original_data.txt")
    return

def check_inspect_clean_data():
    rdf = check_initiate_random_table(csv_name="check_inspect_clean_data.csv")
    df = pd.read_csv(r"check_inspect_clean_data.csv")
    vlab = dv.visual_lab()
    vlab.prepare_for_visual(df)
    vlab.inspect_clean_data( txt_report_filename="check_inspect_clean_data.txt")
    return

def check_inspect_crippled_data():
    rdf = check_initiate_random_table(csv_name="check_inspect_crippled_data.csv")
    df = pd.read_csv(r"check_inspect_crippled_data.csv")
    vlab = dv.visual_lab()
    vlab.prepare_for_visual(df)
    vlab.inspect_crippled_data(txt_report_filename="check_inspect_crippled_data.txt")
    return

def check_generate_level_1_report():
    rdf = check_initiate_random_table(csv_name="check_generate_level_1_report.csv",rate=0.09)
    df = pd.read_csv(r"check_generate_level_1_report.csv")
    vlab = dv.visual_lab()
    vlab.prepare_for_visual(df)
    #### for checking ####
    print(vlab.df)
    print("\nclean df:\n")
    print(vlab.cleanD.clean_df)
    print("\ncrippled df:\n")
    print(vlab.crippledD.crippled_df)
    print("\n\n")
    #######################
    vlab.generate_level_1_report(label_name="gg")
    return

def check_column_module():
    # IN PROGRESS
    # creating random table
    rdf = check_initiate_random_table(csv_name="check_column_module.csv")
    df = pd.read_csv(r"check_column_module.csv")
    cleanD, _, _ = dt.data_sieve(df) # cleanD, crippD, origD
    df= cleanD.clean_df

    # store column in column_module
    # classifications to bool
    my_index = 0
    col_mod = dt.column_module()
    col_mod.get_list_from_df(df, my_index)
    print("column name = ", col_mod.col_name)
    print(col_mod.col_list)
    col_mod.set_conj_discrete_to_bool()
    print(col_mod.col_list_conj)

    # continuous float to scaled
    my_index = 1
    col_mod2 = dt.column_module()
    col_mod2.get_list_from_df(df, my_index)
    print("column name = ", col_mod2.col_name)
    print(col_mod2.col_list)
    col_mod2.set_conj_cont_to_scale()
    print(col_mod2.col_list_conj)

def check_conj_from_discrete_to_bool():
    col=["1","2","3+","1","2"]
    dop=dt.conj_from_discrete_to_bool(col)
    print(dop)

    col = [1, 2, 7, 7, 2, 88, 4, 2, 1, 7, 7, 77, 4]
    col = [str(x) for x in col]
    dop = dt.conj_from_discrete_to_bool(col)
    print(dop)

    col = ["rara", "roror", "meow", "meow", "meow", "roror"]
    col = [str(x) for x in col]
    dop = dt.conj_from_discrete_to_bool(col,drop_one_column=True)
    print(dop)
    return

def check_conj_from_cont_to_scaled():
    col=[0,1,2,6,8,10]
    scale=[0,1]
    out=dt.conj_from_cont_to_scaled(col, scale)
    print(out)

    col = np.random.normal(0,1,36)
    scale = [-1, 1]
    out = dt.conj_from_cont_to_scaled(col, scale)
    print(col)
    print(out)
    ypart=[1.0 for x in col]
    plt.figure()
    plt.scatter(col,ypart)
    plt.figure()
    plt.scatter(out,ypart)
    plt.figure()
    plt.scatter(col,out)
    plt.show()
    return

def check_clean_data():
    rdf = check_initiate_random_table(csv_name="check_clean_data.csv",
                                      rate=0.09,output_label="classification")
    df = pd.read_csv(r"check_clean_data.csv")
    cleanD, _, _ = dt.data_sieve(df)  # cleanD, crippD, origD'
    cleanD.get_list_from_df()
    # for x in cleanD.clean_list:
    #     print(x)

    colname_set=df.columns
    # conversion choices
    # - 1. "discrete_to_bool"
    # - 2. "cont_to_scale"
    # - 3. "discrete_to_int"
    # - 4. ""
    conj_command_set={ colname_set[0]:"discrete_to_bool",
                      colname_set[1]:"cont_to_scale",
                      colname_set[2]:"discrete_to_bool",
                      colname_set[3]:"discrete_to_bool",
                      colname_set[4]:""}
    cleanD.build_column_module_set(conj_command_set)
    for col_mod_key in cleanD.column_module_set:
        print("column:", col_mod_key)
        print(cleanD.column_module_set[col_mod_key].col_list)
        print("conjugate:")
        print(cleanD.column_module_set[col_mod_key].col_list_conj)

    return


def check_build_conj_dataframe():
    rdf = check_initiate_random_table(csv_name="check_build_conj_dataframe.csv",
                                      rate=0.09, output_label="classification")

    df = pd.read_csv(r"check_build_conj_dataframe.csv")
    cleanD, _, _ = dt.data_sieve(df)  # cleanD, crippD, origD'
    cleanD.get_list_from_df()
    colname_set = df.columns

    # conversion choices
    # - 1. "discrete_to_bool"
    # - 2. "cont_to_scale"
    # - 3. "discrete_to_int"
    # - 4. ""
    conj_command_set = {colname_set[0]: "discrete_to_bool",
                        colname_set[1]: "cont_to_scale",
                        colname_set[2]: "discrete_to_bool",
                        colname_set[3]: "discrete_to_bool",
                        colname_set[4]: "discrete_to_int"}
    cleanD.build_conj_dataframe(conj_command_set)

    print(cleanD.clean_df)
    print("\n\nCOMPARE\n\n")
    # print("new set of column names = ",cleanD.colname_set_conj,"\n")
    # count=0
    # for x in cleanD.clean_list_conj:
    #     print(cleanD.colname_set_conj[count],x)
    #     count=count+1
    print(cleanD.clean_df_conj)
    return

def check_build_conj_dataframe2():
    rdf = RDF.RandomDataFrame()
    col1 = {"column_name": "first", "items": [1, 2, 3]}
    itemlist = list(np.linspace(10, 20, 48))
    col2 = {"column_name": "second", "items": itemlist}
    col3 = {"column_name": "third", "items": ["gg", "not"]}
    col4 = {"column_name": "fourth", "items": ["my", "sg", "id", "jp", "us", "bf"]}
    col_out = {"column_name": "result", "items": ["classA", "classB", "classC"]}
    rdf.initiate_random_table(20, col1, col2, col3, col4, col_out, panda=True)
    rdf.crepify_table(rdf.clean_df, rate=0.08)
    rdf.crepified_df.to_csv("testing_conj_df.csv", index=False)

    df = pd.read_csv(r"testing_conj_df.csv")
    cleanD, _, _ = dt.data_sieve(df)  # cleanD, crippD, origD'
    cleanD.get_list_from_df()
    colname_set = df.columns

    # conversion choices
    # - 1. "discrete_to_bool"
    # - 2. "cont_to_scale"
    # - 3. "discrete_to_int"
    # - 4. ""
    conj_command_set = {colname_set[0]: "discrete_to_bool",
                        colname_set[1]: "cont_to_scale",
                        colname_set[2]: "discrete_to_bool",
                        colname_set[3]: "discrete_to_bool",
                        colname_set[4]: "discrete_to_int"}
    discrete_to_int_settings = {"classA": 0, "classB": 1, "classC": 2}
    cont_to_scale_settings = {"scale": [-1, 1], "mode": "uniform"}
    conj_command_setting_set = {colname_set[0]: True,
                                colname_set[1]: cont_to_scale_settings,
                                colname_set[2]: True,
                                colname_set[3]: True,
                                colname_set[4]: discrete_to_int_settings}
    cleanD.build_conj_dataframe(conj_command_set, conj_command_setting_set=conj_command_setting_set)

    print(df)
    print("\n\nCOMPARE : CLEANED\n\n")
    print(cleanD.clean_df)
    print("\n\nCOMPARE : CONJUGATED\n\n")

    print(cleanD.clean_df_conj)
    return

def check_build_conj_dataframe3():
    rdf = RDF.RandomDataFrame()
    # col0 : NOTE THAT IN THIS EXAMPLE we have column for unique ID
    col1 = {"column_name": "first", "items": [1, 2, 3]}
    itemlist = list(np.linspace(10, 20, 48))
    col2 = {"column_name": "second", "items": itemlist}
    col3 = {"column_name": "third", "items": ["gg", "not"]}
    col4 = {"column_name": "fourth", "items": ["my", "sg", "id", "jp", "us", "bf"]}
    col_out = {"column_name": "result", "items": ["classA", "classB", "classC"]}
    rdf.initiate_random_table(20, col1, col2, col3, col4, col_out, panda=True, with_unique_ID="person")
    rdf.crepify_table(rdf.clean_df, rate=0.08)
    rdf.crepified_df.to_csv("testing_conj_df.csv", index=False)

    df = pd.read_csv(r"testing_conj_df.csv")
    cleanD, _, _ = dt.data_sieve(df)  # cleanD, crippD, origD'
    cleanD.get_list_from_df()
    colname_set = df.columns

    # conversion choices
    # - 1. "discrete_to_bool"
    # - 2. "cont_to_scale"
    # - 3. "discrete_to_int"
    # - 4. ""
    conj_command_set = {colname_set[0]: "",
                        colname_set[1]: "discrete_to_bool",
                        colname_set[2]: "cont_to_scale",
                        colname_set[3]: "discrete_to_bool",
                        colname_set[4]: "discrete_to_bool",
                        colname_set[5]: "discrete_to_int"}
    discrete_to_int_settings = {"classA": 0, "classB": 1, "classC": 2}
    cont_to_scale_settings = {"scale": [-1, 1], "mode": "uniform"}
    conj_command_setting_set = {colname_set[0]: None,
                                colname_set[1]: True,
                                colname_set[2]: cont_to_scale_settings,
                                colname_set[3]: True,
                                colname_set[4]: True,
                                colname_set[5]: discrete_to_int_settings}
    cleanD.build_conj_dataframe(conj_command_set, conj_command_setting_set=conj_command_setting_set)

    print(df)
    print("\n\nCOMPARE : CLEANED\n\n")
    print(cleanD.clean_df)
    print("\n\nCOMPARE : CONJUGATED\n\n")

    print(cleanD.clean_df_conj)

    return

def check_build_conj_dataframe4():
    rdf = RDF.RandomDataFrame()
    itemlist = range(100)
    col1 = {"column_name": "first", "items": itemlist}
    col2 = {"column_name": "second", "items": itemlist}
    col3 = {"column_name": "third", "items": itemlist}
    col4 = {"column_name": "fourth", "items": itemlist}
    N_row = 20
    # rdf.initiate_random_table(N_row, col1, col2, col3, col4, panda=True)
    # rdf.clean_df.index = ["".join(('aa', str(x))) for x in range(N_row)]
    rdf.initiate_random_table(N_row, col1, col2, col3, col4, panda=True, row_name_list='aa')
    print(rdf.clean_df)
    return

def check_conj_from_discrete_to_int():
    col = ["a", "b", "c", "d", "a", "b", "d", "e"]
    print(col)
    out = dt.conj_from_discrete_to_int(col)
    print(out)
    newrule = {"a": 100, "b": 200, "c": 7, "d": 1, "e": 2}
    out2 = dt.conj_from_discrete_to_int(col, rule=newrule)
    print("rule = ", newrule)
    print(out2)
    return


def check_get_list_clean():
    gg = [1, 3, np.NaN, 3, 7 , 8, np.NaN]
    gg_fix = dt.get_list_clean(gg)
    print("gg:\n", gg, "\ngg_fix:\n", gg_fix)

    gg2 = ["aa", np.NaN, "cc", "dd", "ff", "gg", np.NaN]
    gg2_fix = dt.get_list_clean(gg2)
    print("gg2:\n", gg2, "\ngg_fix:\n", gg2_fix)
    return

def check_repair_single_column():
    rdf = RDF.RandomDataFrame()
    # col0 : NOTE THAT IN THIS EXAMPLE we have column for unique ID
    col1 = {"column_name": "first", "items": [1, 2, 3]}
    itemlist = list(np.linspace(10, 20, 48))
    col2 = {"column_name": "second", "items": itemlist}
    col3 = {"column_name": "third", "items": ["gg", "not"]}
    col4 = {"column_name": "fourth", "items": ["my", "sg", "id", "jp", "us", "bf"]}
    col_out = {"column_name": "result", "items": ["classA", "classB", "classC"]}
    rdf.initiate_random_table(20, col1, col2, col3, col4, col_out, panda=True, with_unique_ID="person")
    rdf.crepify_table(rdf.clean_df, rate=0.1)
    rdf.crepified_df.to_csv("check_repair_single_column.csv", index=False)

    df = pd.read_csv(r"check_repair_single_column.csv")
    cleanD, _, origD = dt.data_sieve(df)  # cleanD, crippD, origD

    # Choices :
    # 1. mean
    # 2. mean_floor
    # 3. mean_ceil
    # 4. max_occuring
    # 5. min_occuring
    # 6. mid_occuring
    print(origD.original_df)

    origD.initialize_dataframe_repair()
    origD.repair_single_column("first", mode="mean")
    print("\nCOMPARE: REPAIR by 'mean':\n", origD.repaired_df)
    origD.initialize_dataframe_repair()
    origD.repair_single_column("first", mode="mean_floor")
    print("\nCOMPARE: REPAIR by 'mean_floor':\n", origD.repaired_df)
    origD.initialize_dataframe_repair()
    origD.repair_single_column("first", mode="mean_ceil")
    print("\nCOMPARE: REPAIR by 'mean_ceil':\n", origD.repaired_df)

    origD.initialize_dataframe_repair()
    origD.repair_single_column("fourth", mode="max_occuring")
    print("\nCOMPARE: REPAIR by 'max_occuring':\n", origD.repaired_df)
    origD.initialize_dataframe_repair()
    origD.repair_single_column("fourth", mode="min_occuring")
    print("\nCOMPARE: REPAIR by 'min_occuring':\n", origD.repaired_df)
    origD.initialize_dataframe_repair()
    origD.repair_single_column("fourth", mode="mid_occuring")
    print("\nCOMPARE: REPAIR by 'mid_occuring':\n", origD.repaired_df)


def check_repair_table():
    rdf = RDF.RandomDataFrame()
    # col0 : NOTE THAT IN THIS EXAMPLE we have column for unique ID
    col1 = {"column_name": "first", "items": [1, 2, 3]}
    itemlist = list(np.linspace(10, 20, 48))
    col2 = {"column_name": "second", "items": itemlist}
    col3 = {"column_name": "third", "items": ["gg", "not"]}
    col4 = {"column_name": "fourth", "items": ["my", "sg", "id", "jp", "us", "bf"]}
    col_out = {"column_name": "result", "items": ["classA", "classB", "classC"]}
    rdf.initiate_random_table(20, col1, col2, col3, col4, col_out, panda=True, with_unique_ID="person")
    rdf.crepify_table(rdf.clean_df, rate=0.1, column_index_exception=[0]) # do not crepify column 0
    rdf.crepified_df.to_csv("check_repair_single_column.csv", index=False)

    df = pd.read_csv(r"check_repair_single_column.csv")
    cleanD, _, origD = dt.data_sieve(df)  # cleanD, crippD, origD

    # Repair Choices :
    # 1. mean
    # 2. mean_floor
    # 3. mean_ceil
    # 4. max_occuring
    # 5. min_occuring
    # 6. mid_occuring
    repair_setting = {
        "ID": None,
        "first": "mean_floor",
        "second": "mean",
        "third": "max_occuring",
        "fourth": "max_occuring",
        "result": "mid_occuring"
    }
    origD.initialize_dataframe_repair()
    print(origD.original_df, "\n\nCOMPARE: repaired\n")
    origD.repair_table(repair_command_settings=repair_setting)
    print(origD.repaired_df)
    return


def check_ListStats():
    this_list = [1, 2, 4, 777, 4, 3.2, 1, 2, 4, 4, 4, 3.2]
    lstat = gen.ListStats(this_list)
    lstat.prepare_list()

    print("original list = ", this_list)
    print("unique list", lstat.unique_list)
    print("max_occuring = ", lstat.max_occuring)
    print("min_occuring = ", lstat.min_occuring)
    print("mid_occuring = ", lstat.mid_occuring)
    return


def check_generate_level_2_report():
    rdf = check_initiate_random_table(csv_name="check_generate_level_2_report.csv", rate=0.09,
                                      output_label="classification", with_unique_ID=True)
    df = pd.read_csv(r"check_generate_level_2_report.csv")
    vlab = dv.visual_lab()
    vlab.prepare_for_visual(df)
    #### for checking ####
    # print(vlab.df)
    print("\nclean df:\n")
    print(vlab.cleanD.clean_df)
    print("\n\n")
    #######################
    vlab.generate_level_2_report(label_name="rLVL2_")
    return


def check_generate_level_2_report2():
    rdf = check_initiate_random_table(csv_name="check_generate_level_2_report.csv", rate=0.09,
                                      output_label="classification", with_unique_ID=True)
    df = pd.read_csv(r"check_generate_level_2_report.csv")
    vlab = dv.visual_lab()
    vlab.prepare_for_visual(df)
    #### for checking ####
    # print(vlab.df)
    print("\nclean df:\n")
    print(vlab.cleanD.clean_df)
    print("\n\n")
    #######################
    # In this example, analyze against two possible dependent variables 'result' and 'first'
    # against respective independent variables.
    settings = {
        'result': {'first': None, 'second': None, 'third': None, 'fourth': None},
        'first': {'second': 'plot'},
    }
    vlab.generate_level_2_report(label_name="rLVL2_", settings=settings)
    return


def check_multicause_rank():
    rdf = RDF.RandomDataFrame()
    col1 = {"column_name": "first", "items": [1, 2, 3,4]}
    col2 = {"column_name": "second", "items": ["A","B", "C", "D", "E", "F", "G"]}
    col3 = {"column_name": "third", "items": ["gg", "not"]}
    col4 = {"column_name": "fourth", "items": ["my", "sg", "id", "jp", "us", "bf"]}
    col5 = {"column_name": "output", "items": ["classA", "classB", "classC"]}

    rdf.initiate_random_table(1200, col1, col2, col3, col4, col5, panda=True)
    rdf.crepify_table(rdf.clean_df, rate=0.005)
    csv_name = 'check_multicause_rank.csv'
    rdf.crepified_df.to_csv(csv_name, index=False)
    df = pd.read_csv(csv_name)

    vlab = dv.visual_lab()
    vlab.prepare_for_visual(df)
    # print(vlab.cleanD.clean_df)
    indep_var = ["first"]
    dep_var = "output"
    numD1 = vlab.multi_cause_rank(vlab.cleanD.clean_df, indep_var, dep_var)
    numD1.drop_data_points_below(frac=0.9)
    print(numD1.df_number)
    print(numD1.df_frac)
    print("dropped\n", numD1.df_frac_drop)

    indep_var = ["first","second"]
    dep_var = "output"
    numD2= vlab.multi_cause_rank(vlab.cleanD.clean_df, indep_var, dep_var)
    numD2.drop_data_points_below(frac=0.9)
    print(numD2.df_number)
    print(numD2.df_frac)
    print("dropped\n", numD2.df_frac_drop)
    return


def check_drop_data_points_below():
    rdf = RDF.RandomDataFrame()
    itemlist = range(100)
    col1 = {"column_name": "first", "items": itemlist}
    col2 = {"column_name": "second", "items": itemlist}
    col3 = {"column_name": "third", "items": itemlist}
    col4 = {"column_name": "fourth", "items": itemlist}
    N_row = 20
    rdf.initiate_random_table(N_row , col1, col2, col3, col4, panda=True)
    rdf.clean_df.index = ["".join(('aa', str(x))) for x in range(N_row)]

    # vlab = dv.visual_lab()
    # print(rdf.clean_df)
    numD = dv.number_data()
    numD.df_number = rdf.clean_df
    numD.get_frac_from_number()
    print(numD.df_number)
    print(numD.df_frac)
    numD.drop_data_points_below(frac=0.9)
    print("dropping...\n",numD.df_frac_drop)