import numpy as np

def transpose_list(input_list):
    # input_list : n_row by n_col matrix
    # do the usual matrix transpose

    n_col = len(input_list)
    # check valid size
    n_row = len(input_list[0])
    for i in range(n_col):
        if n_row != len(input_list[i]):
            print("list size not compatible for transpose:", n_row, "-vs-",len(input_list[i]))
            return []
    out = []
    for i in range(n_row):
        temp = []
        for j in range(n_col):
            temp.append(input_list[j][i])
        out.append(temp)
    return out


class ListStats():
    def __init__(self, mylist=None):
        self.this_list = mylist
        self.unique_list = None
        self.element_count = None
        self.max_occuring = None  # elements that occur most often
        self.min_occuring = None  # elements that occur least often
        self.mid_occuring = None  # elements that occur mid often
        return

    def prepare_list(self):
        self.unique_list = list(sorted(set(self.this_list), key=self.this_list.index))
        self.max_occuring, self.mid_occuring, self.min_occuring = list_tridiv(self.this_list)
        return

def list_unique_count(data):
    # data : list. not numpy array
    # for example for the data
    # ['a','a','b','c','c','c']
    # this function returns (unique_element,counts)
    # - unique_element = ['a','b','c']
    # - counts = [2,1,3]
    uel = list(sorted(set(data), key=data.index))
    counts=[]
    # for x in uel:
    for i in range(len(uel)):
        counts.append(data.count(uel[i]))
    return uel, counts

def list_all_search_index(data,search_item):
    # data : list. Not numpy array
    # return all indices that correspond to search_item
    # e.g. data=[1,2,1,3,1]
    # if search_item=1, the return indices=[0,2,4]
    indices=[]
    count=0
    temp=data[:]
#     print("start = ",temp)
    while True:
        try:
            search_index=temp.index(search_item)
            indices.append(search_index+count)
            temp.pop(temp.index(search_item))
#             print(" - ",temp)
            count = count + 1
        except:
            break
    return indices
# xx=[1,2,1,3,1,1,1,"a","gg"]
# indices=list_all_search_index(xx,1)
# print(indices)

def list_tridiv(data):
    # data : list. not numpy array
    # given xn the n-th unique element of data
    # and Nxn the number of xn occuring in data
    # - maxD : xn s.t. Nxn = max(Nxn)
    # - midD : xn s.t. Nxn = median(Nxn)
    # - minD : xn s.t. Nxn = min(Nxn)
    uel, counts=list_unique_count(data)
    temp = np.array(counts)
#     print("unique elements\n",uel)
#     print("counts\n",temp)
#     print("sorted counts\n",np.sort(temp))
#     print("list_tridiv(): counts = ", counts)
    counts_median = list(sorted(set(counts), key=counts.index))
    max_count = np.max(counts)
    min_count = np.min(counts)
    mid_count = np.median(counts_median)
#     print("min_count=",min_count)
#     print("mid_count=",mid_count)
#     print("max count = ",max_count)
#     print("temp=",temp)
    maxD = list_all_search_index(counts, max_count)
    midD = list_all_search_index(counts, mid_count)
    minD = list_all_search_index(counts, min_count)
#     print("all max count search\n",maxD)
#     print("all min count search\n",minD)
    if len(midD)==0:
        tempmid=mid_count
        mid_count=np.ceil(tempmid)
        midDright=[]
        while len(midDright) == 0:
            midDright = list_all_search_index(counts,mid_count)
            mid_count = mid_count+1
        midDleft = []
        mid_count = np.floor(tempmid)
        while len(midDleft) == 0:
            midDleft = list_all_search_index(counts,mid_count)
            mid_count = mid_count-1
        midD = list(midDleft)+list(midDright)
    maxD = [uel[x] for x in maxD]
    minD = [uel[x] for x in minD]
    midD = [uel[x] for x in midD]
    # print("list_tridiv(). max occuring = ", maxD)
    # print("list_tridiv(). min occuring = ", minD)
    # print("list_tridiv(). mid occuring = ", midD)
    return maxD, midD, minD



def pair_one_with_list(x, y_set):
    # example
    # x = 1
    # y_set = [2, 3, 5]
    # print(pair_one_with_list(x, y_set))

    # output : [[1, 2], [1, 3], [1, 5]]

    out = []
    for i in range(len(y_set)):
        out.append([x, y_set[i]])
    return out

def pair_list_with_list(x_set, y_set):
    # example
    # x = [1, 2, 3]
    # y = [1, 4, 5]
    # print(pair_list_with_list(x, y))
    # output :
    # [[1, 1], [1, 4], [1, 5], [2, 1], [2, 4], [2, 5], [3, 1], [3, 4], [3, 5]]
    out = []
    for x in x_set:
        temp = pair_one_with_list(x, y_set)
        out = out + temp
    return out

def pair_list_with_LL(x_set, y_LL):
    # example
    # x = [1, 2, 3]
    # y = [1, 4, 5]
    # z= [7,8]
    # xy = pair_list_with_list(x, y)
    # print(pair_list_with_LL(z, xy))

    # output
    # [[7, 1, 1], [7, 1, 4], [7, 1, 5], [7, 2, 1], [7, 2, 4],
    # [7, 2, 5], [7, 3, 1], [7, 3, 4], [7, 3, 5], [8, 1, 1], [8, 1, 4],
    # [8, 1, 5], [8, 2, 1], [8, 2, 4], [8, 2, 5], [8, 3, 1], [8, 3, 4], [8, 3, 5]]
    out = []
    for x in x_set:
        for y_set in y_LL:
            temp = []
            temp.append(x)
            for y in y_set:
                temp.append(y)
            out.append(temp)
    return out

def cartesian_product(full_set, toggle=True):
    # full_set: list of list
    # example
    # x = [1, 2]
    # y = [3, 4]
    # z = [5, 6, 7]
    # k = [88, 99, 100, 200]
    #
    # gg = cartesian_product([x, y, z, k])
    # for x1 in gg:
    #     print(x1)
    N = len(full_set)
    if N == 1:
        if toggle:
            return [[x] for x in full_set[0]]
        else:
            return full_set[0]
    elif N == 2:
        x_set = full_set[0]
        y_set = full_set[1]
        return pair_list_with_list(x_set, y_set)
    elif N > 2:
        return pair_list_with_LL(full_set[0], cartesian_product(full_set[1:],toggle=False))
    return

