import pandas as pd
import numpy as np
import random
from kero.DataHandler.Generic import *

class RandomDataFrame:
    def __init__(self):
        # -- clean_df --
        # store a table, intended to be unaltered, for any purposes

        self.clean_df = None # set by initiate_random_table
        self.crepified_df = None # set by crepify_table
        return

    def initiate_random_table(self, n_row, *argv, panda=True,
                              with_unique_ID=None, row_name_list=None):

        # -- with_unique_ID (string)--
        # If set to None¸then nothing happens.
        # If set to string, add to the first column a unique ID in this manner:
        # - if with_unique_ID="person" then the ID will be person1, person2 etc in order
        #
        # -- n_row (int)--
        # Create a table of n_row rows and as many columns as the
        # number of argv
        #
        # -- *argv --
        # For each k, argv[k] is a dictionary
        # - key : name of column
        # - value : possible data point in the column
        # Each column k is uniform randomly populated with x
        # where x is any one of the element of value.
        #
        # -- panda (Boolean) --
        # True : output tuple (df,[])
        # where df is the random panda DataFrame
        # False: output (x,y) such that x is the matrix that forms
        # a randomly initiated table and y the set of column names
        #
        # example usages:
        # rdf=RandomDataFrame()
        # col1={"column_name": "first", "items":[1,2,3]}
        # itemlist=list(np.linspace(10,20,8))
        # col2={"column_name": "second", "items": itemlist}
        # df,_=rdf.initiate_random_table(4,col1,col2,panda=True)
        # print(df)

        column_name_set=[]
        list_set = []
        if with_unique_ID is not None:
            temp = []
            link = ""
            for i in range(n_row):
                temp.append(link.join((with_unique_ID, str(i))))
            list_set.append(temp)
            column_name_set.append("ID")
        for initset in argv:
            temp = []
            column_name_set.append(initset["column_name"])
            for i in range(n_row):
                item_choice = initset["items"]
                secure_random = random.SystemRandom()
                temp.append(secure_random.choice(item_choice))
            list_set.append(temp)

        # print(column_name_set)
        # print(list_set)
        list_set = transpose_list(list_set)
        df = pd.DataFrame(list_set, columns=column_name_set)
        self.clean_df = df[:]

        if row_name_list is not None:
            self.clean_df.index = ["".join((row_name_list, str(x))) for x in range(n_row)]

        if panda:
            out = (df, [])
        else:
            out = (list_set, column_name_set)
        return out

    def crepify_table(self, dataframe, rate=0.01, column_index_exception=None):
        # Punctuate tables with NaN or empty strings
        # dataframe : panda dataframe
        # rate: fraction between 0 to 1 or None
        # - default value =0.01
        # - if set to None, nothing will happen.
        #   otherwise, for each column whose index is not specified in column_index_exception
        #   the column will be punctuated with blanks at probabilistic rate
        # column_index_exception: index of the column not to be punctured
        # - e.g. [0, 1]
        if rate > 0.1:
            print("crepify_table(): Be reminded that "
                  "larger rate value means more NaN in.\n"
                  "For later processes, this might gives troubles"
                  "due to, for example, empty data frame.")
        df = dataframe[:]
        n_shape = df.shape
        n_row = n_shape[0]
        n_col = n_shape[1]
        # print(n_row,n_col)
        for i in range(n_col):
            if column_index_exception is not None:
                if i in column_index_exception:
                    continue
            for j in range(n_row):
                rand_no = np.random.uniform(0, 1)
                if rand_no > (1-rate):
                    df.at[j, df.columns[i]] = None
        if self.clean_df is not None:
            self.crepified_df = df
        return df

    # Example 1 : initializing data table.
    # rdf = RandomDataFrame()
    # col1 = {"column_name": "first", "items": [1, 2, 3]}
    # itemlist = list(np.linspace(10, 20, 8))
    # col2 = {"column_name": "second", "items": itemlist}
    # col3 = {"column_name": "third", "items": ["gg", "not"]}
    # col4 = {"column_name": "fourth", "items": ["my", "sg", "id", "jp", "us", "bf"]}
    #
    # rdf.initiate_random_table(20, col1, col2, col3, col4, panda=True)
    # print("clean\n", rdf.clean_df)
    #
    # rdf.crepify_table(rdf.clean_df, rate=0.2)
    # print("post crepfify\n", rdf.crepified_df)
    #
    # rdf.crepified_df.to_csv("ggtest.csv", index=False)




