import numpy as np
import cv2, os
from os import listdir
from os.path import isfile, join
from shutil import copyfile


class PhotoBox:
    def __init__(self):
        self.grey = None
        self.grey_center_of_mass = None
        self.grey_sd = None
        self.grey_sd_factor = None
        self.grey_fullweight = None
        return

    def get_greyscale_image(self, grey):
        # Feed it with an array of grey scale image
        # for example, using openCV:
        # gt = cv2.imread(some_file_namep)
        # grey = cv2.cvtColor(gt, cv2.COLOR_BGR2GRAY)
        self.grey = grey
        return

    def get_greyscale_properties(self, sd_power=0.5):
        if self.grey is None:
            print("Use get_greyscale_image() to get grey scale image first")
            return
        img = self.grey
        # print("get_greyscale_center_of_mass(). ")
        # print(img.shape[0],img.shape[1])
        xsum = 0
        ysum = 0
        x2sum = 0
        y2sum = 0
        dim = img.shape
        full_weight = 0
        for y in range(dim[0]):
            for x in range(dim[1]):
                if img[y][x] > 0:
                    # print(x, y, img[y][x])
                    xsum = xsum + x
                    ysum = ysum + y
                    x2sum = x2sum + x**2
                    y2sum = y2sum + y**2
                    full_weight = full_weight + 1
        xcom = xsum/full_weight
        ycom = ysum/full_weight
        xsd = (x2sum/full_weight - xcom**2)**sd_power
        ysd = (y2sum/full_weight - ycom**2)**sd_power
        self.grey_fullweight = full_weight
        self.grey_sd = (int(xsd), int(ysd))
        self.grey_sd_factor = (full_weight/(xsd*ysd))**0.5
        if self.grey_sd_factor > 1:
            self.grey_sd = (int(xsd*self.grey_sd_factor), int(ysd*self.grey_sd_factor))
        self.grey_center_of_mass = (int(xcom), int(ycom))
        return

class GreyScaleWorkShop:
    def __init__(self):
        # print("Building GSWS...")
        return

    def box_positions(self, gsimg):
        # gsimg is a grayscale image
        # color scale : 0 to 1 or any 0 to x scale, for any positive x
        xmin = None
        xmax = None
        ymin = None
        ymax = None
        # cv2.imshow("grey",gsimg)
        # cv2.waitKey(0)
        for i in range(len(gsimg)):
            for j in range(len(gsimg[0])):
                if gsimg[i][j] > 0:
                    if ((xmin is None) or (j < xmin)):
                        xmin = j
                    if ((xmax is None) or (j > xmax)):
                        xmax = j
                    if ((ymin is None) or (i < ymin)):
                        ymin = i 
                    if ((ymax is None) or (i > ymax)):
                        ymax = i
        return xmin, ymin, xmax, ymax

    def scaled_box_positions(self, gsimg, scale="Auto"):
        # color scale : 0 to 1 or any 0 to x scale, for any positive x
        # Scale is either "Auto", None or a tuple (width, height)
        frac = [1, 1]
        width = len(gsimg[0])
        height = len(gsimg)
        if scale is "Auto":
            scale = (200,150)
            # print("Auto...")
            temp = cv2.resize(gsimg, scale)
            frac = [scale[0]/width, scale[1]/height]
            # cv2.imshow("orig", gsimg)
            # cv2.imshow("temp", temp)
        elif scale is None:
            temp = gsimg[:]
        else:
            temp = cv2.resize(gsimg, scale)
            frac = [scale[0]/width, scale[1]/height]

        xmin_s, ymin_s, xmax_s, ymax_s = self.box_positions(temp)
        try:
            xmin = int(xmin_s/frac[0])
            ymin = int(ymin_s/frac[1])
            xmax = int(xmax_s/frac[0])
            ymax = int(ymax_s/frac[1])
        except:
            print(" + scaled_box_positions() returning None")
            xmin=None
            ymin=None
            xmax=None
            ymax=None
        return xmin, ymin, xmax, ymax

    def multiple_scaled_box_positions(self, gsimg, scale="Auto", significant_fraction=0.001):
        # gsimg is 2D matri. Example:
        #   s1 = cv2.imread("pbtest4.png")
        #   s1grey = cv2.cvtColor(s1, cv2.COLOR_BGR2GRAY)/255
        # color scale : 0 to 1 or any 0 to x scale, for any positive x
        # Scale is either "Auto", None or a tuple (width, height)
        # print("multiple_scaled_box_positions(). ")
        frac = [1, 1]
        width = len(gsimg[0])
        height = len(gsimg)
        if scale is "Auto":
            scale = (200,150)
            # print("Auto...")
            temp = cv2.resize(gsimg, scale)
            frac = [scale[0]/width, scale[1]/height]
            # cv2.imshow("orig", gsimg)
            # cv2.imshow("temp", temp)
        elif scale is None:
            temp = gsimg[:]
        else:
            temp = cv2.resize(gsimg, scale)
            frac = [scale[0]/width, scale[1]/height]

        redo = True
        while redo:
            redo = False
            temp = cv2.resize(gsimg, scale)
            component_set=get_connected_components(temp, significant_fraction=significant_fraction)
            rect_set = []
            # print("Size of component_set: ", len(component_set))
            count = 0
            for component in component_set:
                # xmin, ymin, xmax, ymax
                # print("count = ", count)
                count = count + 1
                xmin, ymin, xmax, ymax = self.scaled_box_positions(np.array(component), scale=scale)
                
                if (xmin is None) or (ymin is None) or (xmax is None) or (ymax is None):
                    redo = True
                    break
                else:              
                    xmin = int(xmin/frac[0])
                    ymin = int(ymin/frac[1])
                    xmax = int(xmax/frac[0])
                    ymax = int(ymax/frac[1])
                    if xmin<0:
                        xmin = 0
                    if ymin<0:
                        ymin = 0
                    if xmax>width:
                        xmax = width
                    if ymax>height:
                        ymax = height
                    rect_set.append([xmin, ymin, xmax, ymax])
                        
            if redo:
                if scale is "Auto":
                    new_scale = (300,250)
                elif scale is None:
                    new_scale = "Auto"
                else:
                    new_scale = (scale[0]+100,scale[1]+100)
                print(" + Auto scale change to scale=", new_scale)
                xmin, ymin, xmax, ymax = self.scaled_box_positions( gsimg, scale=new_scale)
                # print(" + new box positions = ", xmin, ymin, xmax, ymax)
                scale = new_scale
                frac = [scale[0]/width, scale[1]/height]
                print("multiple_scaled_box_positions(). Max scale readjusting reached. Consider:")
                print(" + Checking the ground truth image, is there anything missing?")
                print(" + Checking the image itself.")
                if new_scale[0]>width or new_scale[1]> height:
                    print("Terminating...")
                    break
        
        return rect_set

    def extreme_threshold(self,img,thresh=245):
        # only >=253 is retained, the rest is set to 0
        # s1grey = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)/255
        for i in range(len(img)):
            for j in range(len(img[i])):
                is_white=True
                for k in [0,1,2]:
                    if img[i][j][k]<thresh:
                        is_white=False
                if is_white:
                    img[i][j]=[255,255,255]
                else:
                    img[i][j]=[0,0,0]
        s1grey = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        return s1grey
    
    def spawn_ground_truth(self,this_folder, tag_folder,starting_label,
        order_name="img",tag_name="img",
        check_missing_mode=False,rotate_angle_set=None,thresh=254):
        # print("spawn_ground_truth().")
        # print(" - this_folder=",this_folder)
        # print(" - tag_folder=",tag_folder)
        
        tag_rename_clone(this_folder, tag_folder,starting_label,
            order_name=order_name,tag_name=tag_name,filetype=".png",
            clone_name="_CLONE",tag_clone_name="_GT",
            dump_folder="this_dump",
            check_missing_mode=check_missing_mode,
            rotate_angle_set=rotate_angle_set)
        
        folder_to_gt = ''.join((this_folder,"_GT"))
        # print("spawn_ground_truthsfolder_to_gt)
        if not check_missing_mode:
            ALL_FILES = list_all_files_names(folder_to_gt)
            print("spawn_ground_truth().")
            # for x in ALL_FILES:
            #     print(" - ", x)
            totaln = len(ALL_FILES) 
            count = 1
            for filename in ALL_FILES:
                print(" + ", count, " / ", totaln)
                # print(filename)
                fullpath = ''.join((folder_to_gt,"\\",filename))
                s1 = cv2.imread(fullpath)
                et = self.extreme_threshold(s1,thresh=thresh)
                cv2.imwrite(fullpath, et)
                count = count + 1
        return

    def clone_to_annotate(self,this_folder, tag_folder,starting_label,annotation_name,
        order_name="img",tag_name="imggt",
        check_missing_mode=False,rotate_angle_set=None,
        skip_ground_truth=False,
        significant_fraction=0.01,thresh=254,scale="Auto"):

        if check_missing_mode:
            skip_ground_truth=False

        if not skip_ground_truth:
            print("Creating ground truth images...")
            self.spawn_ground_truth(this_folder, tag_folder,starting_label,
                order_name=order_name,tag_name=tag_name,
                check_missing_mode=check_missing_mode,rotate_angle_set=rotate_angle_set,
                thresh=thresh)
        
        if not check_missing_mode:
            print("Starting annotation... Consider using clone_to_annotate_faster() instead.")
            all_gt_imgs = list_all_files_names(''.join((this_folder,"_GT")))
            all_imgs = list_all_files_names(''.join((this_folder,"_CLONE")))
            # for x in all_gt_imgs:
            #     print(x)
            annotation_folder = ''.join((this_folder,"_ANNOT"))
            if not os.path.exists(annotation_folder):
                os.mkdir(annotation_folder)

            for i in range(len(all_gt_imgs)):
                gt = all_gt_imgs[i]
                img_prov = all_imgs[i]
                print(" - ", i+1, "/", len(all_gt_imgs), gt)
                gt3 = cv2.imread(''.join((this_folder,"_GT","\\",gt)))
                imgp3 = cv2.imread(''.join((this_folder,"_CLONE","\\",img_prov)))
                gt_matrix = cv2.cvtColor(gt3, cv2.COLOR_BGR2GRAY)
                width, height=gt_matrix.shape
                # print(gt_matrix.shape)
                # cv2.imshow(str(i), gt_matrix)
                rect_set = self.multiple_scaled_box_positions(gt_matrix/255, scale=scale, significant_fraction=significant_fraction)
                txt = open( ''.join((annotation_folder,"\\",gt[0:len(gt)-4],".txt")) ,'w')
                for k in range(len(rect_set)):
                    txtstr=''.join((annotation_name,"\t",str(width),"\t", str(height)))
                    for j in range(len(rect_set[k])):
                        txtstr=''.join((txtstr,"\t",str(rect_set[k][j])))
                    txtstr=''.join((txtstr,"\n"))
                    # print(txtstr)
                    txt.write(txtstr)
                    # ********* check *********
                    rect = rect_set[k]
                    xmin = rect[0]
                    ymin = rect[1]
                    xmax = rect[2]
                    ymax = rect[3]
                    # cv2.rectangle(gt3, (xmin, ymin), (xmax,ymax),(0,255,0),1)
                    cv2.rectangle(imgp3, (xmin, ymin), (xmax,ymax),(0,255,0),3)
                cv2.imwrite(''.join((annotation_folder,"\\",img_prov)),imgp3)
                txt.close()
        #cv2.waitKey(0)
        return
    def clone_to_annotate_faster(self,this_folder, tag_folder,starting_label,annotation_name,
        order_name="img",
        tag_name="imggt",
        check_missing_mode=False,
        rotate_angle_set=None,
        skip_ground_truth=False,
        significant_fraction=0.01,
        thresh=254,
        scale="Auto",
        discard_edge=True,
        edge_offset=2):

        WARNING1 = "   (!) clone_to_annotate_faster(). Discard annotation by zero dimension."
        WARNING2 = "   (!) clone_to_annotate_faster(). Discard annotation of object at image edge. "

        if check_missing_mode:
            skip_ground_truth=False            

        if not skip_ground_truth:
            print("Creating ground truth images...")
            self.spawn_ground_truth(this_folder, tag_folder,starting_label,
                order_name=order_name,tag_name=tag_name,
                check_missing_mode=check_missing_mode,rotate_angle_set=None,
                thresh=thresh)
        
        if not check_missing_mode:
            print("Starting annotation... Using clone_to_annotate_faster()")
            all_gt_imgs = list_all_files_names(''.join((this_folder,"_GT")))
            all_imgs = list_all_files_names(''.join((this_folder,"_CLONE")))
            # for x in all_gt_imgs:
            #     print(x)
            annotation_folder = ''.join((this_folder,"_ANNOT"))
            if not os.path.exists(annotation_folder):
                os.mkdir(annotation_folder)

            for i in range(len(all_gt_imgs)):
                gt = all_gt_imgs[i]
                img_prov = all_imgs[i]
                print(" - ", i+1, "/", len(all_gt_imgs), gt)
                gt3 = cv2.imread(''.join((this_folder,"_GT","\\",gt)))
                imgp3 = cv2.imread(''.join((this_folder,"_CLONE","\\",img_prov)))
                gt_matrix = cv2.cvtColor(gt3, cv2.COLOR_BGR2GRAY)

                h, w,_ =imgp3.shape
                # print("HW1 = ",h,"/",w)

                # print(gt_matrix.shape)
                # cv2.imshow(str(i), gt_matrix)
                rect_set = self.multiple_scaled_box_positions(gt_matrix/255, scale=scale, significant_fraction=significant_fraction)
                txt = open( ''.join((annotation_folder,"\\",gt[0:len(gt)-4],".txt")) ,'w')
                # print("HW2 = ",h,"/",w)
                for j in range(len(rect_set)):
                    txtstr=''.join((annotation_name,"\t",str(w),"\t", str(h)))
                    for k in range(len(rect_set[j])):
                        txtstr=''.join((txtstr,"\t",str(rect_set[j][k])))
                    txtstr=''.join((txtstr,"\n"))
                    txt.write(txtstr)
                    # ********* check *********
                    rect = rect_set[j]
                    xmin = rect[0]
                    ymin = rect[1]
                    xmax = rect[2]
                    ymax = rect[3]
                    # cv2.rectangle(gt3, (xmin, ymin), (xmax,ymax),(0,255,0),1)
                    cv2.rectangle(imgp3, (xmin, ymin), (xmax,ymax),(0,255,0),1)
                    cv2.imwrite(''.join((annotation_folder,"\\",img_prov)),imgp3)
                txt.close()
                
                current_percent = 0 # add this
                if rotate_angle_set is None:
                    continue
                for i0 in range(len(rotate_angle_set)):
                    angle_in_rad = rotate_angle_set[i0]*np.pi/180
                    angle_in_deg = rotate_angle_set[i0]
                    
                    # annot_rotate_folder="".join((this_folder,"_R",str(i0),"_ANNOT"))
                    # if not os.path.exists(annot_rotate_folder):
                    #     os.mkdir(annot_rotate_folder)
                    # txt = open( ''.join((annot_rotate_folder,"\\",gt[0:len(gt)-4],".txt")) ,'w')
                    txt = open( ''.join((annotation_folder,"\\","R",str(i0),"_",gt[0:len(gt)-4],".txt")) ,'w')
                    
                    img_to_rotate = cv2.imread(''.join((this_folder,"_CLONE","\\",img_prov)))
                    rows,cols,_ = img_to_rotate.shape
                    M = cv2.getRotationMatrix2D((cols/2,rows/2),angle_in_deg,1)
                    dst = cv2.warpAffine(img_to_rotate,M,(cols,rows))
                    
                    for j in range(len(rect_set)):
                        txtstr=''.join((annotation_name,"\t",str(w),"\t", str(h)))
                        rect = rotate_rect(rect_set[j] , angle_in_rad, w,h)
                        for k in range(len(rect)):
                            txtstr=''.join((txtstr,"\t",str(rect[k])))
                        txtstr=''.join((txtstr,"\n"))
                        
                        # ********* check *********
                        xmin = rect[0]
                        ymin = rect[1]
                        xmax = rect[2]
                        ymax = rect[3]

                        do_annot = True

                        if xmin == xmax:
                            print(WARNING1)
                            print("     xmin,ymin,xmax,ymax = ",xmin,ymin,xmax,ymax,end="")
                            print(" | img_height, img_width = ",cols, rows)
                            do_annot = False
                        if ymin == ymax:
                            print(WARNING1)
                            print("     xmin,ymin,xmax,ymax = ",xmin,ymin,xmax,ymax,end="")
                            print(" | img_height, img_width = ",cols, rows)
                            do_annot = False
                        if discard_edge:
                            if xmin <=edge_offset or ymin <= edge_offset or xmax >= cols-edge_offset or ymax >= rows-edge_offset:
                                print(WARNING2)
                                print("     xmin,ymin,xmax,ymax = ",xmin,ymin,xmax,ymax,end="")
                                print(" | img_width, img_height = ",cols, rows)
                                do_annot = False

                        if do_annot:
                            txt.write(txtstr)
                            cv2.rectangle(dst, (xmin, ymin), (xmax,ymax),(0,0,255),3)
                        # print(" + clone_to_annotate_faster(). ", xmin,ymin,xmax,ymax)
                        # cv2.rectangle(gt3, (xmin, ymin), (xmax,ymax),(0,255,0),1)                      
                        # cv2.imwrite(''.join((annot_rotate_folder,"\\",img_prov)),dst)
                        cv2.imwrite(''.join((annotation_folder,"\\","R",str(i0),"_",img_prov)),dst)
                    txt.close()
                    current_percent = time_track(i0,len(rotate_angle_set),current_percent, percent_update=10)


        #cv2.waitKey(0)
        return

def time_track(i, N_total, current_percent, percent_update=1):
    pct = (i+1) / N_total * 100
    if i+1 == N_total:
        print("   ["+"*"*int(pct/10)+'.'*(10-int(pct/10))+"]"+" - DONE")
        return
    if pct >= current_percent:
        print("   ["+"*"*int(pct/10)+'.'*(10-int(pct/10))+"]",end="\r")
        current_percent_next = current_percent + percent_update
    else:
        current_percent_next = current_percent
    # print(current_percent_next)
    return current_percent_next

def get_connected_components(gs_matrix, significant_fraction=0.001):
    # gs_matrix is 2D matrix
    #
    # We distinguish non-zero from zero entries.
    # Connected components are non-zero entries such that
    #   for any two non-zero entries a and b
    #   there is a path connecting a and b.
    # A path is defined as a successive sequence of non-zero entries. For example
    #   0 1 0 0 1
    #   1 1 0 1 1  the matrix has 2 connected components
    #   1 1 0 0 0
    # Return matrix of components
    #   0 1 0 0 0       0 0 0 0 1
    #   1 1 0 0 0  and  0 0 0 1 1
    #   1 1 0 0 0       0 0 0 0 0

    # additional information:
    # - for image processing, convert image to greyscale before using this function
    # - greyscale is sometimes in 0 to 255 scale. Scale it to 0 to 1 first.
    # - output color in 0 to 1 scale as well

    _, segment_set = get_segment_set(gs_matrix)
    # print("*** segment view ***")
    if len(segment_set[0])>0:
        prev_bulk = [[x] for x in segment_set[0]]
    else:
        prev_bulk = [[[0]*len(gs_matrix[0])]]
    # print("initial bulk: ", prev_bulk)
    full_area = None
    if significant_fraction is not None:
        full_area = len(gs_matrix)*len(gs_matrix[0])
    for i in range(1, len(segment_set)):
        # print(" - get_connected_components(). this layer : ", segment_set[i])
        # print("   + get_connected_components(). (before) bulk:\n     ", prev_bulk)
        # print("bulk size = ", len(prev_bulk ), " - index = ", i)
        debug_mode=False
        # if i==97:
        #     debug_mode=True
        prev_bulk = one_step_merge(prev_bulk, segment_set[i], significant_frac=significant_fraction ,full_area=full_area,debug=debug_mode)
        # print("   + get_connected_components(). resulting bulk:\n     ", prev_bulk)
    return prev_bulk

def one_step_merge(prev_bulk, new_layer, significant_frac=None ,full_area=None,
    segment_tolerance=5,
    debug=None):
    if debug:
        print(" one_step_merge(). debug starting...")
        print(" - bulk_size = ", len(prev_bulk))
        print(" - new layer size = ", len(new_layer))
        for i in range(len(new_layer)):
            print("   + ", new_layer[i])
    # (Update)
    # segment_tolerance is in pixel
    # if a segment is found to be very small
    # assume that it MIGHT be noises and remove it

    # merge prev_bulk with a new layer
    # 1 1 1 0 0 0
    # 1 1 1 1 0 1 <=> prev_bulk = [[111000,111100], [000000,000001]]
    # ------------
    # 1 0 0 0 0 1 <=> new_layer = [100000,000001]
    # return  [[111000,111100,100000], [000000,000001,000001]]
    # note that each string 111000 is a list
    # hence prev_bulk are 3 levels of list, new_layer 2 levels

    # example usage:
    # import kero.ImageProcessing.photoBox as kip
    # xx3 = [
    #     [
    #         [1, 1, 0, 0, 1, 1, 0, 0, 0],
    #         [1, 1, 1, 1, 1, 1, 0, 0, 0],
    #         [0, 1, 1, 1, 0, 0, 0, 0, 0],
    #     ],  # face
    #     [
    #         [0, 0, 0, 0, 0, 0, 0, 1, 0],
    #         [0, 0, 0, 0, 0, 0, 0, 1, 1],
    #         [0, 0, 0, 0, 0, 0, 0, 0, 1],  # face
    #     ]
    # ]
    # yy3_1 = [
    #     [0, 0, 1, 1, 0, 0, 0, 0, 0],  # face 
    #     [0, 0, 0, 0, 0, 0, 1, 0, 0],  # face
    #     [0, 0, 0, 0, 0, 0, 0, 0, 1]  # face
    # ]
    # yy3_2 = [
    #     [0, 1, 1, 1, 1, 1, 1, 1, 0],  # face
    # ]
    #
    # print("\nxx3-yy3_1\n", " bulk:", xx3, "\n  new layer:", yy3_1)
    # next_batch31 = kip.one_step_merge(xx3, yy3_1)
    # for x in next_batch31:
    #     print(" + \n ", x)
    # print("\nxx3-yy3_2\n", " bulk:", xx3, "\n  new layer:", yy3_2)
    # next_batch32 = kip.one_step_merge(xx3, yy3_2)
    # for x in next_batch32:
    #     print(" + \n ", x)
    temp = []
    for x in new_layer:
        if sum(x)>segment_tolerance:
            temp.append(x)
    new_layer = temp
    n1 = len(prev_bulk)
    n2 = len(new_layer)
    if n1 == 0:
        return [[x] for x in new_layer]
    if n2 == 0:
        # print(np.array(prev_bulk).shape)
        for i in range(len(prev_bulk)):
            prev_bulk[i].append([0]*len(prev_bulk[0][0]))
        return prev_bulk
    mat_height = len(prev_bulk[0])
    mat_width = len(prev_bulk[0][0])
    # print("one_step_merge()")

    next_batch_layers = []
    full_bulk_index = list(range(len(prev_bulk)))
    for k in range(len(new_layer)):
        bulk_indices_this_layer_merge_with = []
        for j in range(len(prev_bulk)):
            component = prev_bulk[j]
            old_face = component[-1]
            new_face = new_layer[k]
            is_intersect = non_empty_intersection(old_face, new_face)
            # print(" - old face:", old_face)
            # print(" - new face:", new_face)
            # print(" - intersect:", is_intersect)
            if is_intersect:
                # print("  + component: ", component)
                # print("  + new layer[k]: ", new_layer[k])
                bulk_indices_this_layer_merge_with.append(j)

        if len(bulk_indices_this_layer_merge_with) == 0:
            # print(" - An isolated island:")
            temp = [[0]*mat_width]*mat_height
            temp.append(new_layer[k])
            do_merge = False
        else:
            # print(" - Merging ... ")
            temp = [[0] * mat_width] * mat_height
            # print("  + temp (initial) = \n   ", temp)
            for i in bulk_indices_this_layer_merge_with:
                # print("  + prev_bulk[i] = \n   ", prev_bulk[i])
                temp = list_matrix_add(temp, prev_bulk[i])
            # print("  + temp (after) = \n   ", temp)
            # print("  + new_layer [k] = \n   ", new_layer[k])
            temp.append(new_layer[k])
        # print("  + Output:\n   ", temp)
        # print("  + bulk_indices_this_layer_werge_with ", bulk_indices_this_layer_werge_with)

        # (Update here)
        # Either add new bulk or merge
        does_intersect=False
        nth_index=None
        for i in range(len(next_batch_layers)):
            compare=next_batch_layers[i]
            does_intersect = non_empty_intersection(compare[-2], temp[-2])
            if does_intersect:
                nth_index=i
                break
        if does_intersect:
            next_batch_layers[nth_index][-1] = list_add( next_batch_layers[nth_index][-1],temp[-1])
        else:
            next_batch_layers.append(temp)
        full_bulk_index = set_difference(full_bulk_index, bulk_indices_this_layer_merge_with)
    for i in full_bulk_index:
        if (significant_frac is not None) and (full_area is not None):
            frac = 0
            for y in prev_bulk[i]:
                # print("",np.array(y).shape, y[1:100])
                frac = frac + sum(y)/(full_area)
            if frac > significant_frac:
                prev_bulk[i].append([0] * mat_width)
                next_batch_layers.append(prev_bulk[i])
            # print(" + frac / full_area= ", frac, full_area)
        else:
            prev_bulk[i].append([0]*mat_width)
            next_batch_layers.append(prev_bulk[i])
    # print("   + one_step_merge(). DIAGNOSIS 1", np.array(prev_bulk).shape)
    # print("   + one_step_merge(). DIAGNOSIS 2", np.array(next_batch_layers).shape)
    if debug:
        print(" - bulk_size = ", len(next_batch_layers))
        print(" - debug ending...")
    return next_batch_layers

def get_segment_set(mat):
    mat = np.array(mat)
    segment_set = []
    liber_mat_set = []
    temp_mat = mat.copy()
    for k in range(len(temp_mat)):
        segment_set_per_row = []
        liber_mat_set_per_row = []
        while True:
            [this_i, this_j] = find_first_non_zero_index(temp_mat)
            if this_i != k:
                break
            if this_i is None:
                break
            start_entry = (this_i, this_j)
            row_index, column_index_list, temp_mat, liber_mat = find_segment_index(temp_mat, start_entry)
            # print("segment_to_components_at_one_row()")
            # print(" - row, column set: ", row_index, column_index_list)
            # print(" - liber mat: \n", liber_mat)
            # print(" - temp_mat: \n", temp_mat)
            segment_set_per_row.append(liber_mat[k])
            liber_mat_set_per_row.append(liber_mat)
        segment_set.append(segment_set_per_row)
        liber_mat_set.append(liber_mat_set_per_row)
    return liber_mat_set, segment_set

def find_segment_index(mat, start_entry):
    row_index = start_entry[0]
    column_index_list = []
    temp_index = start_entry[1]
    out_mat = mat[:]
    # liber_mat = np.zeros(shape=(len(mat), len(mat[0])))
    liber_mat = [[0]*len(mat[0]) for i in range(len(mat))]
    while True:
        if temp_index == len(mat[0]):
            break
        if mat[row_index][temp_index] == 0:
            break
        else:
            liber_mat[row_index][temp_index] = mat[row_index][temp_index]
            out_mat[row_index][temp_index] = 0
            column_index_list.append(temp_index)
        temp_index = temp_index + 1
    return row_index, column_index_list, out_mat, liber_mat

def find_first_non_zero_index(mat, row_start=0):
    row_indexer = range(len(mat))
    if row_start > 0:
        row_indexer = row_indexer[row_start:]
    for i in row_indexer:
        for j in range(len(mat[i])):
            # print(mat[i][j])
            if mat[i][j] == 0:
                pass
            else:
                return i, j
    return None, None

def list_add(A,B):
    if len(A)!=len(B):
        print("photoBox.py. list_add(). Length A and B not the same")
        return None
    out = []
    for i in range(len(A)):
        out.append(A[i]+B[i])
    return out

def list_matrix_add(A, B):
    # A = [[1,2],[3,4]]
    # B = [[5,5], [6,6]]
    # out = list_matrix_add(A,B)
    # print(out)
    m_dim = len(A)
    n_dim = len(A[0])

    if m_dim != len(B):
        print("list_matrix_add(). dimension does not match. ")
        print(" - length of first/second args = ", m_dim, "/", len(B))
        print(" - arg1 = \n", A, "\n - arg2 = \n", B)
        raise Exception("list_matrix_add(). Error!")
        return None
    if n_dim !=len(B[0]):
        print("list_matrix_add(). dimension does not match")
        print(" - length of 0-th index of first/second args = ", n_dim, "/", len(B[0]))
        print(" - arg1 = \n", A, "\n - arg2 = \n", B)
        raise Exception("list_matrix_add(). Error!")
        return None
    out = [[0]*n_dim for i in range(m_dim)]

    for i in range(n_dim):
        for j in range(m_dim):
            out[j][i] = A[j][i] + B[j][i]
    return out

def non_empty_intersection(seg1, seg2):
    # intersection happens when non-zero entries overlap in the same index
    # [0,0,1] and [0,0,2] intersect
    # [0,1,0] and [1,0,1] do not intersect
    # [1,1,0,0] and [0,1,1,1] intersect
    if len(seg1)!=len(seg2):
        print("non_empty_intersection(). Segments must have the same length!")
        return None
    for i in range(len(seg1)):
        if seg1[i]!=0 and seg2[i]!=0:
            return True
    return False

def set_difference(x, y):
    if len(x)==0:
        return []
    if len(y)==0:
        return x
    temp = x[:]
    for zz in y:
        for i in range(len(temp)):
            if temp[i]==zz:
                try:
                    temp.remove(zz)
                except:
                    pass
                break
    return temp

def linear_bound_fill(in_list):
    # given a list of ones and zeros
    # treat 1 as the bound and 0 otherwise
    # turn all 0 between bounds to 1
    # for example 0000100001000100100 to 0000111111000111100
    
    # pos is either Out, On, Within, OnOut
    x = in_list[:]
    if x[0]==1:
        pos="On"
    else:
        pos="Out"
    for i in range(len(x)):
        if pos is "Out":
            if x[i]==1:
                pos="On"
        elif pos is "On":
            if x[i]==0:
                pos="Within"
                x[i]=1
        elif pos is "Within":
            if x[i]==0:
                x[i]=1
            else:
                pos="OnOut"
        else:
            if x[i]==0:
                pos="Out"
    return x

def list_all_files_names(mypath):
    all_files = [f for f in listdir(mypath) if isfile(join(mypath, f))]
    return all_files

def tag_rename_clone(this_folder, tag_folder,starting_label,
    order_name="img",tag_name="imggt",filetype=".png",
    clone_name="_CLONE",tag_clone_name="_TAG",
    dump_folder="this_dump",
    check_missing_mode=False,
    rotate_angle_set=None):
    # Given this_folder
    # make an exact copy called tag_folder
    # This function makes:
    # 1. a clone of this_folder with all files renamed to order_nameX.filetype where X is a number
    # 2. a clone of tag_folder with all files renamed to tag_nameX.filetype
    #
    # Files in this_folder which somehow are missing their copies in tag_folder
    # will be copied to dump folder
    print("tag_rename_clone().")
    # print(" - this_folder = ", this_folder)

    ALL_FILES = list_all_files_names(this_folder)
    clone_folder = ''.join((this_folder,clone_name))
    tag_clone_folder = ''.join((this_folder,tag_clone_name))
    no_of_files = len(ALL_FILES)
    print("+ Number of files to clone (and rotate) = ", no_of_files)

    df = ''.join((clone_folder,"\\",dump_folder))
    
    if not check_missing_mode:
        if not os.path.exists(clone_folder):
            os.mkdir(clone_folder)
        if not os.path.exists(tag_clone_folder):
            os.mkdir(tag_clone_folder)
        if not os.path.exists(df):
            os.mkdir(df)

    count = starting_label
    no_missing = True
    if not check_missing_mode:
        txt_log = open(this_folder+"_LOG.txt","w")
    for filename in ALL_FILES:
        # print(" - filename",filename)
        fullpath = ''.join((this_folder,"\\",filename))
        tag_filename = ''.join((tag_folder,"\\",filename))
        if not os.path.exists(tag_filename):
            print(" ! missing :", filename)
            no_missing = False
            if not check_missing_mode:
                copyfile(fullpath,''.join((df,"\\",filename)))
        else:            
            if not check_missing_mode:
                if rotate_angle_set is None:
                    copyfile(fullpath,''.join((clone_folder,"\\",order_name,str(count),filetype)))
                    copyfile(tag_filename,''.join((tag_clone_folder,"\\",tag_name,str(count),filetype)))
                    log_string = filename + " -> " +order_name + str(count) + filetype+"\n"
                    txt_log.write(log_string)
                    count = count + 1 
                else:
                    for theta in rotate_angle_set:
                        # print(theta)
                        img = cv2.imread(fullpath)
                        rows,cols,dim = img.shape
                        imgM = cv2.getRotationMatrix2D((cols/2,rows/2),theta,1)
                        imgdst = cv2.warpAffine(img,imgM,(cols,rows))
                        cv2.imwrite(''.join((clone_folder,"\\",order_name,str(count),filetype)),imgdst)

                        img_tag = cv2.imread(tag_filename)
                        rows,cols,dim = img_tag.shape
                        # imgM = cv2.getRotationMatrix2D((cols/2,rows/2),theta,1)
                        imgdst_tag = cv2.warpAffine(img_tag,imgM,(cols,rows))
                        cv2.imwrite(''.join((tag_clone_folder,"\\",tag_name,str(count),filetype)),imgdst_tag)

                        count = count + 1

    if not check_missing_mode:
        txt_log.close()
        DUMP_FILES = list_all_files_names(df)
        if len(DUMP_FILES)==0:
            print(" + All clear! No missing tag files.")
            os.rmdir(df)
    
    if no_missing:
        print("+ ALL GREEN. No missing files.")
    else:
        print("! Check missing files above...")
    return

def rotate_rect(coord, theta, img_width, img_height):
    # coord = [xmin,ymin,xmax,ymax]
    # theta is in rad
    [x01,y01] = rotate_wrt_img_center(coord[0:2], theta, img_width, img_height)
    [x02,y02] = rotate_wrt_img_center(coord[2:4], theta, img_width, img_height)

    # other diagonals
    [x1,y1] = rotate_wrt_img_center([coord[0],coord[3]], theta, img_width, img_height)
    [x2,y2] = rotate_wrt_img_center([coord[2],coord[1]], theta, img_width, img_height)

    xmin0 = min([x01,x02,x1,x2])
    ymin0 = min([y01,y02,y1,y2])
    xmax0 = max([x01,x02,x1,x2])
    ymax0 = max([y01,y02,y1,y2])

    # print(" - rotate_rect : ",xmin0,ymin0,xmax0,ymax0,end="")

    if xmin0<1:
        xmin0 = 1
    if ymin0<1:
        ymin0 = 1
    if xmax0<1:
        xmax0 = 1
    if ymax0<1:
        ymax0 = 1

    if xmin0 > img_width:
        xmin0 = img_width - 1
    if ymin0 > img_height:
        ymin0 = img_height - 1    
    if xmax0 > img_width:
        xmax0 = img_width - 1
    if ymax0 > img_height:
        ymax0 = img_height -1
    # print(" / ",xmin0,ymin0,xmax0,ymax0)
    return [xmin0,ymin0,xmax0,ymax0]

def rotate_wrt_img_center(coord, theta, img_width, img_height):
    # coord = [x,y]
    # theta is in rad
    centercoord = [int(img_width/2),int(img_height/2)]
    vec = [coord[i] - centercoord[i] for i in [0,1]]
    out_vec = [np.cos(theta)*vec[0] + np.sin(theta)*vec[1], -np.sin(theta)*vec[0] + np.cos(theta)*vec[1]]
    out_vec = [out_vec[i] + centercoord[i] for i in [0,1]]
    return [int(x) for x in out_vec]


# copyfile(src, dst)
