import kero.utils.utils as ut
import numpy as np

# print("@NeuralNetwork.py")
# print("@utils.utils test:")
# ut.test()
# print("@@")

class NeuralNetwork:
	def __init__(self):
		print("Initializing a Neural Network object.")
		self.is_consistent = None

		self.weights = None
		self.biases = None	
		
		#----------------------------------------
		# weights : the collection of weights in the neural network
		#   weights is a list [w_l], where w_l is the collection of weights between 
		#     the (l-1)-th and l-th LAYER
		#     for l=2,3,...,L where l=1 is the input layer, l=2 the first hidden layer
		#     and l=L is the output layer
		#   w_l is a matrix (list of list)
		#     so that w_l[i][j] is the weight between neuron j at layer l-1 and 
		#     neuron i at layer l 
		# biases : the collection of biases in the neural network
		#   biases is a list [b_l], where b_l is the collection of biases in the l-th layer
		#     for l=2,3,...,L

		self.number_of_neurons = None # eg [3,3,2,2]
		self.number_of_input = None # number_of_neurons[0]
		self.number_of_output = None # number_of_neurons[-1]
		#----------------------------------------
		# This is set when initiate_neural_network_general() is called
		#
		# number_of_neurons : given L layers, then this is the list 
		#   [N1,N2,...,NL] where Nj is the number of neurons in the j-th layer
		# number_of_input = N1
		# number_of_output = NL

		self.learning_rate=1e-3
		#----------------------------------------
		#
		#

		return

	# NN.L1
	def initiate_neural_network(self,bulk, mode=None,
		verbose = False,
		verbose_init_mode=False,
		verbose_consistency=False):
		if verbose>0 : print(" + Initiate_neural_network(). Mode = ", mode)
		if mode is None:
			weights = bulk["weights"]
			biases = bulk["biases"]
			self.initiate_neural_network_general(weights, biases,
				verbose=verbose_init_mode,
				verbose_consistency=verbose_consistency)
		elif mode=="UniformRandom":
			# number_of_neurons: list of integers [n1,n2,...,nN]
			#   where nk is the number of neurons in layer k
			# 
			number_of_neurons = bulk["number_of_neurons"]
			bounds = bulk["bounds"] # [lower_bound, upper_bound] 
			weights = []
			biases = []

			# optional initiation option
			# layerwise_normalization: Bool. If true, divide the random uniform distribution
			#    by the number of neurons in the previous layer.
			layerwise_normalization = False
			if "layerwise_normalization" in bulk:
				layerwise_normalization = bulk["layerwise_normalization"]
				if verbose>40:
					print("   layerwise_normalization activated.")

			for i in range(1,len(number_of_neurons)):
				layerwise_NORM = 1
				if layerwise_normalization:
					layerwise_NORM = 1/number_of_neurons[i-1] 
				if verbose>40:
					print("     layer: ",i-1,"to",i,"layerwise_NORM = ",layerwise_NORM)
					print("       (-) no of neurons at layer i-1 = ", number_of_neurons[i-1] )
				w_l = layerwise_NORM*np.random.uniform(low=bounds[0],high=bounds[1],size=(number_of_neurons[i],number_of_neurons[i-1]))
				b_l = [0]*number_of_neurons[i]
				# b_l = np.random.uniform(low=bounds[0],high=bounds[1],size=(number_of_neurons[i],1))
				weights.append(ut.numpy_matrix_to_list(w_l))
				biases.append(b_l)
			self.initiate_neural_network_general(weights, biases,
				verbose=verbose_init_mode,
				verbose_consistency=verbose_consistency)
		return

	# NN.L2
	def initiate_neural_network_general(self,weights, biases,
		verbose=False,
		verbose_consistency=False):
		if verbose>0 : print(" -+ initiate_neural_network_general().")
		# ---------------------------------------
		# Initialize weights and biases manually
		# weights : see the format in __init__(self)
		# biases : see the format in __init__(self)
		is_consistent, number_of_neurons = self.neuron_number_consistency_check(weights, biases,
			verbose=verbose_consistency)
		self.number_of_neurons = number_of_neurons
		self.number_of_input = number_of_neurons[0]
		self.number_of_output = number_of_neurons[-1]
		self.is_consistent = is_consistent

		self.weights = weights
		self.biases = biases

		if verbose>30:
			print("    self.weights:")
			for w_l in self.weights:
				ut.print_numpy_matrix(np.matrix(w_l),formatting="%10.4f",no_of_space=12)
				print("                    -------------------------")
			print("    self.biases:")
			for b_l in self.biases:
				ut.print_numpy_matrix(np.transpose(np.matrix(b_l)),formatting="%10.4f",no_of_space=12)
				print("                    -------------------------")
		return

	# NN.L3
	def neuron_number_consistency_check(self, weights, biases, verbose=False):
		# ---------------------------------------
		# weights : see the format in __init__(self)
		# biases : see the format in __init__(self)
		# return
		#   is_consistent : bool, if true, then the inputs are compatible as Neural Network
		#   number_of_neurons : list [n1,n2,...nL], number of neurons in layer 1 to L,
		#     where layer 1 is the input layer and layer L the output layer
		is_consistent = True
		number_of_neurons = []
		if verbose>0: print(" --+ neuron_number_consistency_check().") 
		n_layer = 2
		for w_l in weights:
			is_matrix_bool, row_size, col_size = ut.list_is_matrix(w_l)
			if verbose>10: print("     layer: ", n_layer-1, " to ", n_layer)
			if verbose>10: print("     is_matrix/row/col = ", is_matrix_bool,"/", row_size,"/", col_size)
			if not is_matrix_bool:
				print("     inconsistent weight matrix found [1]. Exiting with:")
				return False, []
			if n_layer == 2:
				number_of_neurons = number_of_neurons + [col_size, row_size]
			else:
				if number_of_neurons[-1] != col_size:
					print("     inconsistent weight matrix found [2]. Exiting.")
					if verbose>10: print("     ", number_of_neurons," || ", col_size)
					return False, []
				number_of_neurons = number_of_neurons + [row_size]
			n_layer = n_layer + 1
		number_of_neurons_2 = []
		for b_l in biases:
			number_of_neurons_2 = number_of_neurons_2 + [len(b_l)]
		if len(number_of_neurons_2)+1==len(number_of_neurons):
			for j in range(1,len(number_of_neurons)):
				if number_of_neurons[j] != number_of_neurons_2[j-1]:
					print("     inconsistent weight-bias dimensions [2]. Exiting.")
					if verbose>10: print("     1. ", number_of_neurons)
					if verbose>10: print("     2. ", number_of_neurons_2)
					return False, []
		else:
			print("     inconsistent weight-bias dimensions [1]. Exiting.")
			if verbose>10: print("     1. ", number_of_neurons)
			if verbose>10: print("     2. ", number_of_neurons_2)
			return False, []
		return is_consistent, number_of_neurons
	# NN.Ex.L1
	def print_neural_network_information(self,verbosity=1):
		print(" > print_neural_network_verbose_information().")
		# is_consistent, number_of_neurons=self.neuron_number_consistency_check(self.weights, self.biases, verbose=False)
		 
		print("   Dimension consistent? :", self.is_consistent)
		print("   Layers? [Input, 2,3,..., Output] :", self.number_of_neurons)
		if verbosity>5: print("    -> no of input/output = ",self.number_of_input,"/", self.number_of_output)
		print("   Learning rate : ",self.learning_rate)
		if verbosity>0:
			# TODO: add more information
			pass
		return
	# NN.Ex.L2


class NetworkUpdater():
	def __init__(self):
		self.current_weights = None # (list, not in numpy matrix)
		self.current_biases = None # (list)
		self.learning_rate = None # (float)
		self.number_of_neurons = None # list [N1,N2,...,NL]
		self.number_of_input = None # number_of_neurons[0]
		self.number_of_output = None # number_of_neurons[-1]
		self.bias_drag = 1e-2
		#------------------------------------------------------------
		# These properties are the same as that of NeuralNetwork

		self.input_set = None 
		self.Y0_set = None 
		#------------------------------------------------------------
		# input_set :
		# - list of numpy matrix [x]. Each x a column vector m x 1, m the number of neurons in input layer.
		# Y0_set :
		# - list of numpy matrix [Y0]. Each Y0 nx1, where n is the no of neurons in layer l=L
		#   The true/observed values in the output layer corresponding to the input set.
		#   In another words, for each k=1,...,N, Y0_set[k] = f(x[k]) where f is the true function that
		#   our neural network is modelling and N the number of data points.

		self.method = "RegularStochastic"
		self.method_specific_settings = None # (dictionary)
		
		return

	# NU.L1

	def update_wb(self, input_set, Y0_set, weights, biases, AF,	
		mse_mode="compute_only",	
		verbose=False):

		#----------------------------------------
		# weights : the collection of weights in the neural network
		#   weights is a list [w_l], where w_l is the collection of weights between 
		#     the (l-1)-th and l-th layer
		#     for l=2,3,...,L where l=1 is the input layer, l=2 the first hidden layer
		#     and l=L is the output layer
		#   w_l is a matrix (list of list)
		#     so that w_l[i][j] is the weight between neuron j at layer l-1 and 
		#     neuron i at layer l 
		# biases : the collection of biases in the neural network
		#   biases is a list [b_l], where b_l is the collection of biases in the l-th layer
		#     for l=2,3,...,L
		# input_set : list of numpy matrix [x]. Each x a column vector m x 1, m the number of neurons in input layer  
		# AF : activationFunction object

		# mse_mode
		# None: do not compute mse at all
		# "compute_only" : 
		# "compute_and_print" : 

		# back-propagation
		# Assume the following have been called:
		#   set_training_data()
		#   set_neural_network()
		#   set_settings()

		print(" + update_wb().")
		mse_list = []
		
		if mse_mode is not None:
			Y_set = []
			n = len(input_set)
			for j in range(n): # for each data points
				a_l_set, _ = self.feed_forward( weights, biases, input_set[j], AF,
									verbose=False,
									matrix_formatting="%6.2f")
				Y_set.append(a_l_set[-1])
			current_mse = MSE(Y_set,Y0_set)
			mse_list.append(current_mse)
			if mse_mode is "compute_and_print":	
				print("   Intital mse = ",current_mse)	

		weights_next = weights[:]
		biases_next = biases[:]

		if self.method=="RegularStochastic":
			batch_size = self.method_specific_settings["batch_size"]
			no_of_epoch = self.method_specific_settings["no_of_epoch"]
			do_shuffle = self.method_specific_settings["shuffle_batch"]
			if verbose>20:
				print("   batch size = ",batch_size)
				print("   no of epoch = ", no_of_epoch)
				print("   shuffle batch = ", do_shuffle)
			
			for epoch_i in range(no_of_epoch):
				# if verbose>20: print("   epoch : ", epoch_i+1)
				collection_of_batches, index_partition = ut.partition_list(input_set, batch_size, do_shuffle=do_shuffle )
				partition_of_Y0_set, _ = ut.partition_list(Y0_set, index_partition)

				if verbose>30:
					for xx,yy in zip(collection_of_batches, partition_of_Y0_set):
						# 	print(xx,"\n",yy)
						print("   * partition size consistency check: ", len(xx)," and ",len(yy))
						print("   * --------------------------------------")
					
				for i in range(len(collection_of_batches)):
					# Recall: each batch (list of numpy matrix) is a subset of input_set
					input_set_batch = collection_of_batches[i] # is input_set - like
					Y0_set_batch = partition_of_Y0_set[i]
					if verbose>40: 
						print("   batch ", i, " : ")
					#print("   input_set_batch :\n",input_set_batch)
					
					for x in input_set_batch:
						# Recall: each x (numpy matrix) is an element of input_set
						if verbose>30:
							ut.print_numpy_matrix(x,formatting="%6.2f",no_of_space=10)
							print("         -------")
					weights_next, biases_next = self.update_wb_regular_stochastic(input_set_batch, Y0_set_batch,
						weights_next, biases_next, AF)

				# for accuracy measurement
				# Y0_set :list of numpy matrix over all data points
				if mse_mode is not None:
					Y_set_next = []
					n = len(input_set)
					for j in range(n): # for each data points
						a_l_set_next, _ = self.feed_forward( weights_next, biases_next, input_set[j], AF,
											verbose=False,
											matrix_formatting="%6.2f")
						Y_set_next.append(a_l_set_next[-1])
					current_mse = MSE(Y_set_next,Y0_set)
					mse_list.append(current_mse)
					if mse_mode is "compute_and_print":	
						print("   epoch :",epoch_i+1," - MSE = ",current_mse)

		return weights_next, biases_next, mse_list

	def set_settings(self, method="RegularStochastic",
		method_specific_settings={"batch_size":4,"no_of_epoch":1}):
		# List all available methods here
		#
		# method = "RegularStochastic"
		# method_specific_settings:
		#   batch_size: (integer)
		#   no_of_epoch: (integer)
		#   shuffle_batch : (bool)
		#------------------------------------------------------------
		#   Given n data points. Partition to m = ceil(n/batch_size) batches
		#   For every update wb_n to wb_(n+1), average the contribution to gradient descent
		#   for all batches
		#	In this implementation, 1 epoch is 1 update
		self.method = method
		self.method_specific_settings = method_specific_settings
		return

	def set_training_data(self,input_set,Y0_set):
		# See __init__() for arguments explanations
		self.input_set = input_set
		self.Y0_set = Y0_set
		return

	def set_neural_network(self, NeuralNetwork):
		# NeuralNetwork (NeuralNetwork): assume initiated, with weights and biases
		#   i.e. assume  initiate_neural_network_general() has been called successfully
		self.current_weights = NeuralNetwork.weights
		self.current_biases = NeuralNetwork.biases	

		self.learning_rate = NeuralNetwork.learning_rate

		self.number_of_neurons = NeuralNetwork.number_of_neurons
		self.number_of_input = NeuralNetwork.number_of_input
		self.number_of_output = NeuralNetwork.number_of_output
		return

	# NU.L2

	def update_wb_regular_stochastic(self, input_set, Y0_set,
		weights, biases, AF,		
		verbose=False,
		verbose_feed_forward=False,
		verbose_compute_diff=False,
		verbose_delta_L=False,
		verbose_compute_delta=False):
		if verbose>0: print(" -+ update_wb_regular_stochastic().")
		#----------------------------------------
		# weights : the collection of weights in the neural network
		#   weights is a list [w_l], where w_l is the collection of weights between 
		#     the (l-1)-th and l-th layer
		#     for l=2,3,...,L where l=1 is the input layer, l=2 the first hidden layer
		#     and l=L is the output layer
		#   w_l is a matrix (list of list)
		#     so that w_l[i][j] is the weight between neuron j at layer l-1 and 
		#     neuron i at layer l 
		# biases : the collection of biases in the neural network
		#   biases is a list [b_l], where b_l is the collection of biases in the l-th layer
		#     for l=2,3,...,L
		# input_set :list of numpy matrix [x]. Each x a column vector m x 1, m the number of neurons in input layer
		# Y0_set :list of numpy matrix over all data points
		# AF : activationFunction object

		n = len(input_set)
		L = len(weights) + 1
		backprop_label = range(2,L+1)
		backprop_label = [ backprop_label[-1-i] for i in range(len(backprop_label))]
		#-------------------------------------------------
		# for example if there are L=6 layers, this will be [6,5,4,3,2]

		collection_a_l_set = [] # over all data points
		collection_z_l_set = []
		for j in range(n): # for each data points
			a_1 = input_set[j]
			a_l_set, z_l_set = self.feed_forward( weights, biases, a_1, AF,
							verbose=verbose_feed_forward,
							matrix_formatting="%6.2f")
			collection_a_l_set.append(a_l_set)
			collection_z_l_set.append(z_l_set) # collection of z_l_set over all data points
		Y_set = [ a_l_set[-1] for a_l_set in collection_a_l_set]

		weights_next = [] # list [w_2,...,w_L] where each w_k is a list of list a matrix
		biases_next = []

		for i in range(len(backprop_label)): # for each layer
			# for regular_stochastic, at each l, compute average over data points
			# note that weights = [w_2,w_3,...,w_L] so the corresponding index [0,1,..,l-2,...,L-2]
			l = backprop_label[i]
			max_i = len(backprop_label)-1
			if l == L:
				if verbose>10:
					print("    Layer (Output): ",backprop_label[i], " || i = ", i,"/",max_i)
				average_dterm_at_L, average_delta_at_L , CDLADP , CdLADP = self.compute_averaged_dterm_for_L(
					n, input_set, Y_set, Y0_set, AF,
					collection_z_l_set, collection_a_l_set,
					verbose_delta_L=verbose_delta_L,
					verbose_compute_diff=verbose_compute_diff)
				# average_dterm_at_l (numpy matrix)
				CDlp1ADP = CDLADP
				Cdlp1ADP = CdLADP
				w_l = np.matrix(weights[-1])
				w_l_next = w_l - self.learning_rate * average_dterm_at_L
				b_l = np.transpose(np.matrix(biases[-1]))
				b_l_next = b_l - self.bias_drag * self.learning_rate * average_delta_at_L
			else:
				if verbose>10:
					print("    Layer: ",backprop_label[i], " || i = ", i,"/",max_i)
				w_l_plus_1 = weights[l-2 + 1]
				average_dterm_at_l, average_delta_at_l , CDlADP , CdlADP = self.compute_averaged_dterm_for_other_l( 
					l, n, w_l_plus_1, Cdlp1ADP, collection_a_l_set, collection_z_l_set, AF,
					verbose=False,
					verbose_compute_diff=verbose_compute_diff,
					verbose_compute_delta=verbose_compute_delta)
				# average_delta_at_l  (numpy matrix)
				CDlp1ADP = CDlADP
				Cdlp1ADP = CdlADP
		
				w_l = np.matrix(weights[-1 - i])
				w_l_next = w_l - self.learning_rate * average_dterm_at_l
				b_l = np.transpose(np.matrix(biases[-1-i]))
				b_l_next = b_l - self.bias_drag * self.learning_rate * average_delta_at_l
			weights_next.insert(0, ut.numpy_matrix_to_list(w_l_next))
			biases_next.insert(0, ut.numpy_matrix_to_list(b_l_next))

		return weights_next, biases_next

	def compute_differential_term_at_l_per_data_point(self, delta_l, a_l_minus_1,
		verbose=False):
		# differential terms hre defined as delta^l times transpose of a^(l-1)
		#   delta_l (numpy matrix): vector of size m x 1, where m is no of neurons at layer l
		#   a_l_minus_1 (numpy matrix)  x 1 vector, signals from layer l-1
		if verbose>0: print(" -+ compute_differential_term_at_l().")
		dterm = np.matmul(delta_l,np.transpose(a_l_minus_1))
		# dterm (numpy matrix)
		return dterm

	def compute_delta_l_per_data_point(self, w_l_plus_1, delta_l_plus_1, z_l, AF,
		verbose=False,
		print_format="%6.8f"):
		if verbose>0: print(" -+ compute_differential_term_at_l().")
		# recursive formula to compute d_l for all l except l=L
		#-----------------------------------------------
		# w_l_plus_1 (list): matrix of size m x n, weights between layer l and l+1
		# delta_l_plus_1 (numpy matrix): vector of size m x 1, where m is no of neurons at layer l+1
		# z_l (numpy matrix): vector of size m x 1, m no of neurons at layer m
		#   this is w_l . a_l_minus_1 + b_l
		# return
		#  delta_l (numpy matrix): vector of size m x 1, where m is no of neurons at layer l
		if verbose>5: 
			print("    np.array(w_l_plus_1).shape = ",np.array(w_l_plus_1).shape)
		temp = np.matmul(np.transpose(np.matrix(w_l_plus_1)),delta_l_plus_1)
		temp2 = [AF.afp(z_l.item(i)) for i in range(len(z_l))]
		if not (len(temp)==len(temp2)):
			print(" -+ compute_delta_l(). Error, inconsistent dimensions. return None.")
			return None
		delta_l = []
		for i in range(len(temp)):
			delta_l.append(temp.item(i)*temp2[i])
		delta_l = np.transpose(np.matrix(delta_l))
		if verbose>10:
			print("    delta_l = ") 
			ut.print_numpy_matrix(delta_l,formatting=print_format,no_of_space=10)
		return delta_l

	def compute_delta_L_per_data_point(self, z_L, Y0, Y_set, Y0_set, AF,
		verbose=False,
		print_format="%6.8f"):	
		if verbose>0: print(" -+ compute_delta_L_per_data_point()")	

		# a single data point:
		#   z_L (numpy matrix) is a vector: output data computed from NN before activation
		#   Y0 (numpy matrix) is a vector: observed/true output data. 
		# Still, we are minimizing MSE over the entire set of data points
		#   Y_set and Y0_set are respectively the collection of NN-computed and true/observed data points
		#
		# Assume a_L and Y0 have the same size
		# Assume Y_set and Y0_set have the same size
		# Assume all entries of a_L and Y0 are real
		# AF (activationFunction). Assume it is initiated.
		# return
		#  delta_L (numpy matrix)
		
		nabla_mse = nabla_MSE(Y_set,Y0_set) 
		#-----------------------------------
		# (numpy matrix) a vector of size m, m the number of neurons in layer L
		temp = [AF.afp(z_L.item(i)) for i in range(len(z_L))]
		#-----------------------------------
		# 

		delta_L = []
		if not (len(nabla_mse)==len(temp)):
			raise ValueError(" -+ compute_delta_L(). Error, inconsistent dimensions. Return None.")
			return None
		for i in range(len(nabla_mse)):
			delta_L.append(nabla_mse.item(i)*temp[i])
		delta_L = np.transpose(np.matrix(delta_L))
		if verbose>10:  
			print("    delta_L = ") 
			ut.print_numpy_matrix(delta_L,formatting=print_format,no_of_space=10)
		return delta_L

	# NU.L3

	def feed_forward(self, weights, biases, a_1, AF,
		verbose=False,
		matrix_formatting="%6.2f"):
		if verbose>0: print(" --+ feed_forward()")
		#----------------------------------------
		# weights : the collection of weights in the neural network
		#   weights is a list [w_l], where w_l is the collection of weights between 
		#     the (l-1)-th and l-th layer
		#     for l=2,3,...,L where l=1 is the input layer, l=2 the first hidden layer
		#     and l=L is the output layer
		#   w_l is a matrix (list of list)
		#     so that w_l[i][j] is the weight between neuron j at layer l-1 and 
		#     neuron i at layer l 
		# biases : the collection of biases in the neural network
		#   biases is a list [b_l], where b_l is the collection of biases in the l-th layer
		#     for l=2,3,...,L
		# a_1 : (numpy matrix) input layer 
		# AF (activationFunction). Assume it is initiated.

		# matrix_print_raw_style (bool). If true, then matrix is printed as is
		#   if false, then matrix is printed with padding.

		# return
		# a_l_set
		#   (list) [a_l] for l = 1, 2, ..., L, [l=1 is input, l=L is output]
		#   each a_l (numpy matrix) m x 1 vector, where m is no of neurons in layer l
		# z_l_set
		#   (list) [z_l] for l = 1, 2, ..., L, [l=1 is input, l=L is output]
		#   each z_l (numpy matrix) m x 1 vector, where m is no of neurons in layer l
		# 
		# Note: each entry a_l[k] is just the corresponding entry z_l[k] with
		#   activation function applied
		# Note: a_1 == z_1

		a_l_set = [a_1]
		z_l_set = [a_1]
		N_layer_minus_1 = len(weights)
		if not (N_layer_minus_1 == len(biases)):
			print("     Error.Dimension of weights and biases inconsistent.")
		for i in range(N_layer_minus_1):
			a_l_minus_1 = a_l_set[-1]
			w_l = np.matrix(weights[i])
			b_l = np.transpose(np.matrix(biases[i]))
			if verbose>20: 
				print("     ------------------------------------")
				print("     layer ",i, "to layer", i+1)
				print("     w_l = " )
				ut.print_numpy_matrix(w_l,formatting=matrix_formatting,no_of_space=20)
				print("     a_l_minus_1 = ")
				ut.print_numpy_matrix(a_l_minus_1,formatting=matrix_formatting,no_of_space=20)
				print("     b_l = ")
				ut.print_numpy_matrix(b_l,formatting=matrix_formatting,no_of_space=20)
			z_l = np.matmul(w_l,a_l_minus_1) + b_l
			z_l_set.append(z_l)
			a_l_act = []
			for i in range(len(z_l)):
				if verbose>30 : print("      -> ",i, " : ", z_l.item(i), " : ",AF.af(z_l.item(i)))
				a_l_act.append(AF.af(z_l.item(i)))
			a_l_act = np.transpose(np.matrix(a_l_act))	
			if verbose>20 : 
				print("     a_l_act = ") 
				ut.print_numpy_matrix(a_l_act,formatting=matrix_formatting,no_of_space=20)
			a_l_set.append(a_l_act)
			
		return a_l_set, z_l_set

	# NU.L4

	def compute_z_j_at_layer_l(self, a_l_minus_1, w_l, b_l, 
		verbose=False):
		if verbose>0: print(" ---+ compute_z_j_at_layer_l().")
		# w_L (numpy matrix) is a matrix, weights between layers L-1 and L
		# b_L (numpy matrix) is a vector, biases at layer L
		# a_L_minus_1 (numpy matrix) is a vector, signal from layer L-1
		z_j_at_layer_l = np.matmul(w_l,a_l_minus_1) + b_L
		return z_j_at_layer_l

	def compute_averaged_dterm_for_other_l(self, l, n, w_l_plus_1, Cdlp1ADP, 
		collection_a_l_set, collection_z_l_set, AF,
		verbose=False,
		verbose_compute_diff=False,
		verbose_compute_delta=False):

		# dterm is just a convenient name for update terms for w. Effectively this
		#   function computes the update term for l as well.

		# l (int): for update in layer l<L
		# n (int) : no of data points

		# Cdlp1ADP
		#   collection_of_delta_at_l_plus_1_over_all_data_points
		#   delta_l_plus_1 (numpy matrix): vector of size m x 1, where m is no of neurons at layer l+1

		# collection_z_l_set
		#   collection of z_l_set over all data points
		# collection_a_l_set
		#   collection of a_l_set over all data points
		# AF : activationFunction object
		# w_l_plus_1 is a matrix (list of list)
		#     so that w_l[i][j] is the weight between neuron j at layer l-1 and 
		#     neuron i at layer l 

		CDlADP = [] 
		# collection_of_DTERM_at_l_over_all_data_points
		#   dterm (numpy matrix): matrix of size m x n, where m is no of neurons at layer l, n at layer l-1
		#   n (int) : no of data points
		CdlADP = []
		# collection_of_delta_at_l_over_all_data_points
		#   delta_l (numpy matrix): vector of size m x 1, where m is no of neurons at layer l
		#   n (int) : no of data points
		
		for j in range(n): # for each data points
			delta_l_plus_1 = Cdlp1ADP[j]
			a_l_set = collection_a_l_set[j]
			z_l_set = collection_z_l_set[j]
			# a_l_set
			#   (list) [a_l] for l = 1, 2, ..., L, [l=1 is input, l=L is output]
			#   each a_l (numpy matrix) m x 1 vector, where m is no of neurons in layer l
			# z_l_set
			#   (list) [z_l] for l = 1, 2, ..., L, [l=1 is input, l=L is output]
			#   each z_l (numpy matrix) m x 1 vector, where m is no of neurons in layer l
			z_l = z_l_set[l-1]
			a_l_minus_1 = a_l_set[l- 1 - 1] 
			delta_l = self.compute_delta_l_per_data_point(w_l_plus_1, delta_l_plus_1, z_l, AF,
				verbose=verbose_compute_delta,
				print_format="%6.8f")
			# delta_l (numpy matrix): vector of size m x 1, where m is no of neurons at layer l
			
			dterm_per_datapoint = self.compute_differential_term_at_l_per_data_point(
				delta_l, a_l_minus_1,
				verbose=verbose_compute_diff)
			# dterm (numpy matrix)

			CDlADP.append(dterm_per_datapoint)
			CdlADP.append(delta_l)

			if j == 0:
				average_dterm_at_l = 1/n*dterm_per_datapoint
				average_delta_at_l = 1/n*delta_l
			else:
				average_dterm_at_l = average_dterm_at_l + 1/n*dterm_per_datapoint
				average_delta_at_l = average_delta_at_l + 1/n*delta_l
			
			# average_dterm_at_l (numpy matrix), dimension of matrix of weight w_l 
			# average_delta_at_l (numpy matrix), dimension of column vector
			

		return average_dterm_at_l, average_delta_at_l , CDlADP , CdlADP

	def compute_averaged_dterm_for_L(self, n, input_set, Y_set, Y0_set, AF,
		collection_z_l_set, collection_a_l_set,
		verbose_delta_L=False,
		verbose_compute_diff=False):

		# dterm is just a convenient name for update terms for w. Effectively this
		#   function computes the update term for l=L as well.

		# collection_z_l_set
		#   collection of z_l_set over all data points
		# collection_a_l_set
		#   collection of a_l_set over all data points

		CDLADP = [] 
		# collection_of_DTERM_at_L_over_all_data_points
		#   dterm (numpy matrix): matrix of size m x n, where m is no of neurons at layer l, n at layer l-1
		#   n (int) : no of data points
		CdLADP = []
		# collection_of_delta_at_L_over_all_data_points
		#   delta_l (numpy matrix): vector of size m x 1, where m is no of neurons at layer l
		#   n (int) : no of data points

		for j in range(n): # for each data points			
			#-----------------------------------------------
			# a_l_set
			#   (list) [a_l] for l = 1, 2, ..., L, [l=1 is input, l=L is output]
			#   each a_l (numpy matrix) m x 1 vector, where m is no of neurons in layer l
			# z_l_set
			#   (list) [z_l] for l = 1, 2, ..., L, [l=1 is input, l=L is output]
			#   each z_l (numpy matrix) m x 1 vector, where m is no of neurons in layer l
			# 
			# Note: each entry a_l[k] is just the corresponding entry z_l[k] with
			#   activation function applied
			# Note: a_1 == z_1

			a_1 = input_set[j] # a_1 : (numpy matrix) input layer 
			Y0 = Y0_set[j]
			z_l_set = collection_z_l_set[j]
			z_L = z_l_set[-1]
			a_l_set = collection_a_l_set[j]
			a_l_minus_1 = a_l_set[- 1 - 1] 
			delta_L = self.compute_delta_L_per_data_point(z_L, Y0, Y_set, Y0_set, AF,
				verbose=verbose_delta_L,
				print_format="%6.8f")
			delta_l = delta_L
			dterm_per_datapoint = self.compute_differential_term_at_l_per_data_point(
					delta_l, a_l_minus_1,
					verbose=verbose_compute_diff)

			CDLADP.append(dterm_per_datapoint)
			CdLADP.append(delta_L)

			if j == 0:
				average_dterm_at_L = 1/n*dterm_per_datapoint
				average_delta_at_L = 1/n*delta_L
			else:
				average_dterm_at_L = average_dterm_at_L + 1/n*dterm_per_datapoint
				average_delta_at_L = average_delta_at_L + 1/n*delta_L
		return average_dterm_at_L, average_delta_at_L , CDLADP , CdLADP

	# NN.Ex.L1
	def print_all_network_udpater_parameter_level_1(self):
		print(" > print_neural_network_verbose_information().")
		print("   Current weights size = ", len(self.current_weights) ) # (list, not in numpy matrix)
		print("   Current biases size = ", len(self.current_biases) ) # (list)
		print("   learning_rate = ",self.learning_rate )# (float)
		print("   number_of_neurons = ",self.number_of_neurons ) # list [N1,N2,...,NL]
		print("   number_of_input (no of neurons in input layer) = ",self.number_of_input ) # number_of_neurons[0]
		print("   number_of_output (no of neurons in output layer) = ",self.number_of_output) # number_of_neurons[-1]
		print("   size of input_set = ",len(self.input_set) )
		print("   size of Y0_set = ",len(self.Y0_set) )
		print("   method = ",self.method )
		print("   method_specific_settings = ",self.method_specific_settings) # (dictionary)
		return
		

class activationFunction:
	def __init__(self, func = "Sigmoid"):
		self.af = None # activation function
		self.afp = None # first derivative of activation function, p for prime
		self.select_activation_function(func)
		return

	def select_activation_function(self, func):
		if func == "Sigmoid":
			self.af = lambda x: 1/(1+np.exp(-x))
			self.afp = lambda x: 1/(1+np.exp(-x)) * (1- 1/(1+np.exp(-x)))
		else:
			# In any other exceptional cases, set to Sigmoid function
			self.af = lambda x: 1/(1+np.exp(-x))
			self.afp = lambda x: 1/(1+np.exp(-x)) * (1- 1/(1+np.exp(-x)))
		return


# Static functions
def MSE(Y_set,Y0_set):
	# Y_set is a list of Y
	# Y0_set is a list of Y0
	# Assume Y_set and Y0_set have the same size, otherwise error will be raised
	#
	# Y (numpy matrix) is a vector: computed data from model
	# Y0 (numpy matrix) is a vector: observed/true data
	# Assume Y and Y0 have the same size, otherwise error will be raised
	# Assume all entries are real
	
	mse = 0
	n = len(Y_set)
	if not (n==len(Y0_set)):
		print("MSE(). Error. Inconsistent dimension between Y_set and Y_set0. Return nothing.")
		return	
	for i in range(n):
		Y = Y_set[i]
		Y0 = Y0_set[i]
		Nout1 = len(Y.tolist())
		Nout2 = len(Y0.tolist())
		if not (Nout1==Nout2):
			print("MSE(). Error. Inconsistent dimension between Y and Y0. Return nothing.")
			return
		# now compute the norms between true and model output values
		for j in range(Nout1):
			mse = mse + (np.abs(Y0.item(j)-Y.item(j)))**2
	mse = mse / (2*n)

	return mse

def nabla_MSE(Y_set,Y0_set):
	# Y_set is a list of Y
	# Y0_set is a list of Y0
	# Assume Y_set and Y0_set have the same size, otherwise error will be raised
	#
	# gradient of MSE wrt a_L
	# Y (numpy matrix) is a column vector: computed output from model
	# Y0 (numpy matrix) is a column vector: observed/true data
	# Assume Y and Y0 have the same size, otherwise error will be raised
	# Assume all entries are real
	n = len(Y_set)
	if not (n==len(Y0_set)):	
		print(" >> Y_set size = ",len(Y_set))
		print(" >> Y0_set size = ",len(Y0_set))
		raise ValueError("nabla_MSE(). Error. Inconsistent dimension between Y_set and Y_set0. Return nothing.")
		return
	nabla_mse = None
	for i in range(n):
		Y = Y_set[i] # numpy matrix
		Y0 = Y0_set[i]
		if i == 0 :
			nabla_mse = Y-Y0
		else:
			nabla_mse = nabla_mse + Y-Y0
	nabla_mse = nabla_mse/n 
	return nabla_mse