from mpl_toolkits.mplot3d import Axes3D  
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation

class Network:
	def __init__(self):
		# member_set is a dictionary
		# - key : each member should be string identified 
		# - values : each member's value is the set of other members' key
		# - if x : { y, ... } then y: { x, ... }
		self.member_set = {} 

		# grid_lattice_points is a dictionary
		# - key : (string) when initialized, it has the same key as member_set
		# - values : tuples showing the key's coordinates in space, typically (x,y,z) or 3D 
		self.grid_lattice_points = None

		# set_of_fields is a dictionary
		# - key : (string) name of the field
		# - values : any fields. 
		# For each key, the value is intended to be a dictionary with 
		#   + key : (string) member of the network
		#   + values : the value of the field, can be scalar, vector etc
		# However, this field can be anything 
		self.set_of_fields = {}

	def add_new_field(self,new_field):
		return

	def add_member(self,new_member):
		# new_member : string, key of dictionary
		self.member_set[new_member] = []
		return

	def add_link(self,member1,member2):
		# member1, member2 : strings, keys of dictionary
		# print(self.member_set[member1],self.member_set[member2])
		if member1 in self.member_set and member2 in self.member_set:
			self.member_set[member1].append(member2)
			self.member_set[member2].append(member1)
		else:
			if not (member1 in self.member_set):
				print(" + add_link(). Member with id ", member1, " does not exist.")
			if not (member2 in self.member_set):
				print(" + add_link(). Member with id ", member2, " does not exist.")
		return

	def assign_lattice_points(self, grid_lattice_points):
		# grid coordinates are 3D (x,y,z)
		# - dictionary. key = member identifier (string)
		#               value = (x,y,z)
		
		# check that the members of grid_lattice_points are equal
		#  to the members of the network
		grid_identifiers = [x for x in grid_lattice_points]
		network_identifiers = [x for x in self.member_set]

		# print(" - grid keys: ", grid_identifiers)
		# print(" - network keys: ",network_identifiers)
		key_matching = grid_identifiers == network_identifiers
		if key_matching:
			print(" - Key matching. identical? ", key_matching)
			self.grid_lattice_points = grid_lattice_points
		else:
			print(" - Failed key matching. Not plotting graph.")
			print("   Exactly each point in the network needs a (x,y,z) coordinates.")
			print("   No more, no less.")
		return

	def visualize_3D_graph(self, ax,
						no_of_points_per_edge=24,
						title = None,
						# subplot_index=111,
						this_figure=None,
						member_set=None,
						color_set=None,
						cmap=None,
						xlim=None,
						ylim=None,
						zlim=None,
						):
		# member_set: dictionary
		# - key : string
		# - values : list of strings, connected members
		if self.grid_lattice_points is None:
			print(" - visualize_3D_graph(). Lattice points not assigned.")
			return
		if this_figure is None:
			fig = plt.figure()
		else:
			fig = this_figure

		# pro='3d'
		# if plot2d:
		# 	pro = None
		X = [self.grid_lattice_points[x][0] for x in self.grid_lattice_points]
		Y = [self.grid_lattice_points[x][1] for x in self.grid_lattice_points]

		xint = range(int(min(X)), int(np.ceil(max(X))+1))
		ax.set_xticks(xint)
		yint = range(int(min(Y)), int(np.ceil(max(Y))+1))
		ax.set_yticks(yint)

		if ax.name is '3d':
			ax.name = '3d'
			Z = [self.grid_lattice_points[x][2] for x in self.grid_lattice_points]
			p = ax.scatter(X, Y, Z, c=color_set,cmap=cmap)
			if member_set is not None:
				for i, txt in enumerate([key for key in member_set]):
					ax.text(X[i],Y[i],Z[i],str(txt))

			# Drawing edge #######
			for akey in self.member_set:
				# print(" + key = ", akey, " : ", self.member_set[akey])
				point1 = self.grid_lattice_points[akey]
				# print(" + anchor point: ",point1)
				for akey_linked in self.member_set[akey]:
					point2 = self.grid_lattice_points[akey_linked]
					# print("   linked point: ",point2)
					edge = edge_maker(point1,point2, no_of_points_per_edge)
					X1 = [edge[i][0] for i in range(len(edge))]
					Y1 = [edge[i][1] for i in range(len(edge))] 
					Z1 = [edge[i][2] for i in range(len(edge))] 
					ax.scatter(X1, Y1, Z1, c='r', s=1)
			##########################
			zint = range(int(min(Z)), int(np.ceil(max(Z))+1))
			ax.set_zticks(zint)
			if zlim is not None:
				ax.set_zlim(zlim[0],zlim[1])
			# ax.set_xlabel('X Label')
			# ax.set_ylabel('Y Label')
			# ax.set_zlabel('Z Label')
		else:
			p = ax.scatter(X, Y, c=color_set,edgecolors='k',cmap=cmap)
			if member_set is not None:
				for i, txt in enumerate([key for key in member_set]):
					ax.annotate(txt,(X[i],Y[i]))

			# Drawing edge #######
			for akey in self.member_set:
				# print(" + key = ", akey, " : ", self.member_set[akey])
				point1 = self.grid_lattice_points[akey]
				# print(" + anchor point: ",point1)
				for akey_linked in self.member_set[akey]:
					point2 = self.grid_lattice_points[akey_linked]
					# print("   linked point: ",point2)
					edge = edge_maker(point1,point2, no_of_points_per_edge)
					X1 = [edge[i][0] for i in range(len(edge))]
					Y1 = [edge[i][1] for i in range(len(edge))] 
					ax.scatter(X1, Y1, c='r', s=1)
			##########################
		if xlim is not None:
			ax.set_xlim(xlim[0],xlim[1])
		if ylim is not None:
			ax.set_ylim(ylim[0],ylim[1])

		# ax.set_xlabel('X Label')
		# ax.set_ylabel('Y Label')
		if title is not None:
			plt.title(title)
		if cmap is not None:
			plt.colorbar(p)
		return

	def build_1D_chain(self,N):
		lattice_x_coord = 0
		self.grid_lattice_points={}
		for num_id in range(N):
			member_id = str(num_id+1)
			self.add_member(member_id)
			self.grid_lattice_points[member_id] = [lattice_x_coord,0,0]
			lattice_x_coord = lattice_x_coord + 1
		for i in range(N-1):
			if i == 0:
				self.add_link(str(i+1),str(i+2))
			else:
				self.add_link(str(i+1), str(i+2))
				# self.add_link(str(i+1), str(i))
		return

	def build_2D_grid(self,m,n):
		# m row (y axis)
		# n columns (x axis)
		lattice_y_coord = 0
		lattice_x_coord = 0
		self.grid_lattice_points={}

		for num_id in range(m*n):
			member_id = str(num_id+1)
			self.add_member(member_id)
			self.grid_lattice_points[member_id] = [lattice_x_coord,lattice_y_coord,0]
			lattice_x_coord = lattice_x_coord + 1
			if lattice_x_coord==n:
				lattice_x_coord = 0
				lattice_y_coord = lattice_y_coord+1

		tuple_to_linear = {}
		linear_to_tuple = {}
		# temp=0
		for i in range(m*n):
			col_index = int(i%n)
			row_index = int((i-col_index)/n)
			tuple_to_linear[(row_index,col_index)] = i
			linear_to_tuple[i]=(row_index,col_index)
			# if col_index != temp:	
			# 	print("")
			# print("(",row_index,",",col_index,")",",", end="")
			# temp=col_index
		for i in range(m*n):
			this_coord = linear_to_tuple[i]
			N = coord_add(this_coord,(0,1),1,1)
			S = coord_add(this_coord,(0,1),1,-1)
			E = coord_add(this_coord,(1,0),1,1)
			W = coord_add(this_coord,(1,0),1,-1)
			for x in [N,S,E,W]:
				if x[0]>=0 and x[0]<m and x[1]>=0 and x[1]<n:
					neighbor = str(tuple_to_linear[x] + 1)
					self.member_set[str(i+1)].append(neighbor)
		return

	def attach_a_simple_probability_vector_field(self):
		self.set_of_fields["prob"] = {}
		action_set = ["a1","a2"]
		action_index_set = [0,1] # numerrical labelling:  a1=0, a2=1
		for v in self.member_set:
			# a1 and a2 are actions. This probability vector field is created such that
			# at a time step, each member has 50% chance of doing a1 and 50% of doing a2
			self.set_of_fields["prob"][v] = [action_set,[0.5,0.5]]	
		return action_set, action_index_set

class Network_processor:
	def __init__(self):
		self.member_set=None
		self.action_history=None
		self.field_history=None
		# Tset is specified in terms of range(N)
		self.Tset=None

		self.action_to_index_map={}
		self.index_to_action_map={}

		return

	def map_action_index(self,action_set,action_index_set):
		for ac,ind in zip(action_set,action_index_set):
			self.action_to_index_map[ac]=ind
			self.index_to_action_map[ind]=ac
		return
	
	def add_member_and_history(self,member_set,Tset,action_history):
		self.member_set=member_set
		self.action_history=action_history
		self.Tset=Tset
		return
	
	def add_new_field_history(self, field_history):
		self.field_history=field_history
		# for field in field_history:
		# 	self.field_history[field] = field_history[field]
		return

	def compute_interval_mean(self, mean_over_N=1):
		# mean_over_N: integer. Take an interval from time t-mean_over_N to t,
		#  where t is time now. Compute the mean value of number of each action
		#  within the interval [t-mean_over_N,t] for every time t
		interval_mean_history={}
		for v in self.member_set:
			interval_mean_history[v]=[]
		for t in self.Tset[1:]:
			# Tset[1:] is used to avoid error of writing range(0,0)
			if t<mean_over_N:
				interval_start=0
				interval_end=t
			else:
				interval_start=t-mean_over_N
				interval_end=t

			for v in self.member_set:
				action_index_set = [x[0] for x in self.action_history[v]]
				# print(action_index_set)
				current_interval = action_index_set[interval_start:interval_end]
				# print(current_interval)
				n_temp = len(current_interval)
				frac_set=[]
				for this_index in self.index_to_action_map:
					temp=count_no_of_elements(this_index,action_index_set,interval_start,interval_end)
					frac=temp/n_temp
					# print(" - index=", this_index, " : ", temp, " -> ",frac)
					frac_set.append(frac)
				interval_mean_history[v].append(frac_set)
		# e.g. output
		# each dictionary value is time varying field. In this example, is a 2D vector field, 10 time steps
		# {
		# '1': [[1.0, 0.0], [0.5, 0.5], [0.6666666666666666, 0.3333333333333333], [0.5, 0.5], [0.6, 0.4], [0.6666666666666666, 0.3333333333333333], [0.7142857142857143, 0.2857142857142857], [0.75, 0.25], [0.6666666666666666, 0.3333333333333333]], 
		# '2': [[1.0, 0.0], [0.5, 0.5], [0.6666666666666666, 0.3333333333333333], [0.75, 0.25], [0.6, 0.4], [0.6666666666666666, 0.3333333333333333], [0.7142857142857143, 0.2857142857142857], [0.75, 0.25], [0.6666666666666666, 0.3333333333333333]], 
		# '3': [[0.0, 1.0], [0.5, 0.5], [0.3333333333333333, 0.6666666666666666], [0.5, 0.5], [0.4, 0.6], [0.3333333333333333, 0.6666666666666666], [0.42857142857142855, 0.5714285714285714], [0.5, 0.5], [0.4444444444444444, 0.5555555555555556]], 
		# '4': [[1.0, 0.0], [1.0, 0.0], [0.6666666666666666, 0.3333333333333333], [0.5, 0.5], [0.4, 0.6], [0.5, 0.5], [0.42857142857142855, 0.5714285714285714], [0.375, 0.625], [0.4444444444444444, 0.5555555555555556]]
		# }
		return interval_mean_history

	def extract_interval_mean_at_a_time_step(self,interval_mean_history,time_step=None):
		if time_step is None:
			time_step=-1
		interval_mean_at_t = {}
		for i in self.index_to_action_map:
			interval_mean_at_t[i]={}
			for v in interval_mean_history:
				interval_mean_at_t[i][v]=[]
			for v in interval_mean_history:
				# v is member
				interval_mean_at_t[i][v].append(interval_mean_history[v][time_step][i])
		# e.g. output
		#  {0: {'1': [0.25], '2': [0.5], '3': [0.5], '4': [0.25]}, 1: {'1': [0.75], '2': [0.5], '3': [0.5], '4': [0.75]}}
		# outer dictionary key is the index of action; inner dictionary key is the member of network
		return interval_mean_at_t

	def plot_interval_mean_over_time(self, 
		ax,
		interval_mean_history,
		member_set_to_display,
		action_index_set_to_display,
		color_set=["b","r","g","c","m","y","k"],
		style_set=["-","-.","--",":"],
		plotting_interval=(1,None),
		plotting_interval_step_length=None,
		title=None):
		# plotting_interval is set as a tuple of (1,M) where M<=N
		#   given self.Tset = range(N)
		
		color_count=0
		for v in member_set_to_display:
			style_count=0	
			for i in action_index_set_to_display:
				# print(style_count%len(action_index_all),style_count%len(net1D.member_set))
				action_i_over_time = [a[i] for a in interval_mean_history[v]]
				if plotting_interval[1] is None:
					t = self.Tset[plotting_interval[0]:]
					a = action_i_over_time[plotting_interval[0]-1:]
					if plotting_interval_step_length is not None:
						temp = t
						t = [temp[i] for i in range(len(temp)) if i%plotting_interval_step_length==0]
						temp = a
						a = [temp[i] for i in range(len(temp)) if i%plotting_interval_step_length==0]
					plt.plot(t, 
						a,
						linestyle=style_set[style_count%len(style_set)],
						color=color_set[color_count%len(color_set)],
						label=str(v)+":"+self.index_to_action_map[i])
				else:
					t = self.Tset[plotting_interval[0]:plotting_interval[1]]
					a = action_i_over_time[plotting_interval[0]:plotting_interval[1]]
					if plotting_interval_step_length is not None:
						temp = t
						t = [temp[i] for i in range(len(temp)) if i%plotting_interval_step_length==0]
						temp = a
						a = [temp[i] for i in range(len(temp)) if i%plotting_interval_step_length==0]
					plt.plot(t,
					    a,
						linestyle=style_set[style_count%len(style_set)],
						color=color_set[color_count%len(color_set)],
						label=str(v)+":"+self.index_to_action_map[i])
				style_count=style_count+1
			color_count=color_count+1
		plt.title(title)
		ax.legend()
		return

	def plot_action_over_time(self,
		ax,
		member_set_to_display=None,
		plot_interval=(0,None),
		color_set=["b","r","g","c","m","y","k"],
		title=None
		):

		# member_set_to_display
		# - set members of network whose action over time is plotted
		# - if set to default None, then plot all members
		if member_set_to_display is None:
			member_set_to_display = self.member_set

		style_count=0
		color_count=0
		
		maxz=None
		minz=None
		for v in member_set_to_display:
			this_action_index_set = [x[0] for x in self.action_history[v]]

			if plot_interval[1] is None:	
				Z=this_action_index_set[plot_interval[0]:]	
				if ax.name=='3d':
					ax.plot(self.Tset[plot_interval[0]:],		
						[style_count]*len(self.Tset[plot_interval[0]:]),
						Z,
						markersize=4,label=v,
						c=color_set[color_count%len(color_set)])
				else:
					ax.plot(self.Tset[plot_interval[0]:],Z,		
						markersize=4,label=v,
						c=color_set[color_count%len(color_set)])
			else:
				Z=this_action_index_set[plot_interval[0]:plot_interval[1]]
				if ax.name=='3d':	
					ax.plot(self.Tset[plot_interval[0]:plot_interval[1]],
						[style_count]*len(self.Tset[plot_interval[0]:plot_interval[1]]),
						Z,		
						markersize=4,label=v,
						c=color_set[color_count%len(color_set)])
				else:
					ax.plot(self.Tset[plot_interval[0]:plot_interval[1]],Z,		
						markersize=4,label=v,
						c=color_set[color_count%len(color_set)])
			if minz is None:
				minz=min(Z)
			if maxz is None:
				maxz=max(Z)
			if maxz<max(Z):
				maxz=max(Z)
			if minz>min(Z):
				minz=min(Z)
			color_count=color_count+1
			style_count=style_count+1
		ax.set_xlabel('X, Time step')
		
		if ax.name=='3d':
			zint = range(int(minz),int(maxz)+1)
			ax.set_zlabel('Z, Action-index')
			ax.set_zticks(zint)
			ax.set_ylabel('Y, members')
			yint = range(0,style_count)
			ax.set_yticks(yint)
		else:
			yint = range(int(minz),int(maxz)+1)
			ax.set_ylabel('Y, Action-index')
			ax.set_yticks(yint)
		ax.legend()
		plt.title(title)
		return

class SAVE_POINT():
	def __init__(self):
		self.data={}
		return

def edge_maker(point1,point2, no_of_points):
	# point1, point2 are (x,y,z)
	#   return set of points (x,y,z) such that they fill a straightline between
	#   point1 and point2.
	one_vec = coord_add(point1,point2,-1/(no_of_points+1),1/(no_of_points+1))
	return [(point1[0]+i*one_vec[0],point1[1]+i*one_vec[1],point1[2]+i*one_vec[2]) for i in range(1,no_of_points+1)]

def coord_add(point1,point2,c1,c2):
	# point1, point2 are (x,y,z), (x1,y1,z1)
	# c1, c2 are scalars
	# return (c1*x+c2*x1, c1*y+c2*y1, c1*z+c2*z1)
	if len(point1)==3 and len(point2)==3:
		return (c1*point1[0]+c2*point2[0],c1*point1[1]+c2*point2[1],c1*point1[2]+c2*point2[2])
	elif len(point1)==2 and len(point2)==2:
		return (c1*point1[0]+c2*point2[0],c1*point1[1]+c2*point2[1])
	else:
		print(" - coord_add(). Coord addition not defined for the given inputs.")
	return None 

def choose_action_continuous_probability(action_set, probability_set, N=1):
	# action_set: list of strings [a1,a2,...,aN]
	# probability_set: list of float [p1,p2,...,pN]. 
	# + probability set will be normalized to p1+...+pN=1
	# N: positive integer. number of actions to be chosen randomly
	if not (len(action_set)==len(probability_set)):
		print(" - choose_action_continuous_probability(). Each action set must have exactly 1 weight.")
		return None

	for p in probability_set:
		if p<0:
			print(" - choose_action_continuous_probability(). Probability_set must contain positive values.")
			return None
	
	partition=[]
	total=0
	for i in range(len(probability_set)):
		total = total + probability_set[i]
		partition.append(total) 
	for i in range(len(partition)):
		partition[i]=partition[i]/total
	# print(partition)
	action_out=[]
	action_index=[]
	for i in range(N):
		nth = 0
		coin = np.random.uniform()
		while True:
			if coin <= partition[nth]:
				action_out.append(action_set[nth])
				action_index.append(nth)
				break
			else:
				nth = nth + 1
	if N==1:
		return action_out[0], nth
	return action_out, action_index

def count_no_of_elements(x,this_set,start_index,end_index):
	# this_set is any list
	# return number of x in this_set within index start (inclusive) and end (exclusive)
	count = 0
	for i in range(start_index,end_index):
		if x==this_set[i]:
			count = count + 1
	return count 

def tuple_to_grid(tlist,reshape=None):
	# tlist is a list of list of 2 or 3, i.e. only 2D or 3D
	# e.g. [[1,1],[2,1],[11,11]] or
	#      [[1,1,2],[2,1,4],[11,11,10]]
	# reshape = [m,n,p] integers, the dimension of output matrix
	# - if set to None, output is 1D list
	X=[]
	Y=[]
	Z=None
	is3D=False
	if len(tlist[0])==3:
		is3D=True
		Z=[]
	for el in tlist:
		X.append(el[0])
		Y.append(el[1])
		if is3D:
			Z.append(el[2])

	if reshape is not None:
		X = np.reshape(X,(reshape[0],reshape[1]))
		Y = np.reshape(Y,(reshape[0],reshape[1]))
		if is3D:
			Z= np.reshape(Z,(reshape[0],reshape[1]))
			Z=Z.tolist()
	return X.tolist(),Y.tolist(),Z

def grid_to_tuple(X,Y,Z=None):
	out=[]
	for i in range(len(X)):
		for j in range(len(X[0])):
			# print(i,j)
			if Z is None:
				out.append([X[i][j],Y[i][j]])
			else:
				
				out.append([X[i][j],Y[i][j],Z[i][j]])
	return out

class SpecialGallery:
	# This class is created with the sole purpose of 
	#   plotting specialized figures and storing data
	#   w.r.t the figures
	def __init__(self):
		self.quiver3D_lattice3D_settings={}
		return

	def quiver3D_lattice3D(self,
		fig,
		ax,
		no_of_steps_to_animate,
		grid_lattice_points,
		field_history,
		field_name_to_plot,
		field_index_to_plot=[0,1,2],
		field_name_label=["X","Y","Z"],
		grid_shape=None,
		time_between_steps_in_ms=1,
		gif_filename='quiver.gif',
		show_real_time=True,
		view_angle_elevation=90,
		view_angle_azimuth=90,
		xlim=None,
		ylim=None,
		zlim=None,
		arrow_length_scale=None,
		):
		# This plot 3D quiver 
		# - lattice is specified in 3D
		# - field chosen from field history may be more or equal to 3D.
		#   However, field_index_to_plot

		# field_history: dictionary
		#   key: string, field name
		#   value: dictionary, key=member name, value= fields values (context dependent)
		# field_name_to_plot: string
		#   out of all the fields stored in field_history, choose the one to plot via this argument
		# field_index_to_plot: list of integers [i,j,k]
		#   out of the fields chosen, if the field is more than 3D, select the 3 indices to display
		#   i.e. the 3D subspace of the field
		lattice_in_tuple_list = [grid_lattice_points[x] for x in grid_lattice_points]
		if grid_shape is None:
			grid_shape = [len(lattice_in_tuple_list), 1]
		X,Y,Z = tuple_to_grid(lattice_in_tuple_list,reshape=grid_shape)

		if xlim is None:
			xlim = [min(min(X))-1,max(max(X))+1]
		if ylim is None:
			ylim = [min(min(Y))-1,max(max(Y))+1]
		if zlim is None:
			zlim = [min(min(Z))-1,max(max(Z))+3]

		probability_grid_history=[]
		for i in range(len(field_history)):
		    field_now = field_history[i] 
		    probability_field_in_tuple_list = [field_now['prob'][x][1] for x in field_now['prob']] # ASSUME every list of available actions are the same
		    U,V,W = tuple_to_grid(probability_field_in_tuple_list,reshape=grid_shape)
		    probability_grid_history.append([U,V,W])

		U,V,W = None,None,None
		Q = ax.scatter(0,0)

		def update_quiver(num, Q, X, Y, W, ax):
		    field_now = field_history[num]
		    tfil = [field_now[field_name_to_plot][x][1] for x in field_now[field_name_to_plot]] # ASSUME every list of available actions are the same
		    probability_field_in_tuple_list=[]
		    for tf in tfil:
		    	temp=[]
		    	for i in field_index_to_plot:
		    		temp.append(tf[i])
		    	probability_field_in_tuple_list.append(temp)
		    # print(probability_field_in_tuple_list)
		    # print("----------------")
		    U,V,W = tuple_to_grid(probability_field_in_tuple_list,reshape=grid_shape)
		    if arrow_length_scale is not None:
		    	U = np.array(U)*arrow_length_scale
		    	V = np.array(V)*arrow_length_scale
		    	W = np.array(W)*arrow_length_scale
		    ax.cla()
		    Q = ax.quiver(X, Y, Z, U, V, W, color='r')
		    ax.set_xlim(xlim[0],xlim[1]) # ax.set_xlim(-1, 4)
		    ax.set_ylim(ylim[0],ylim[1]) # ax.set_ylim(-1, 4)
		    ax.set_zlim(zlim[0],zlim[1]) # ax.set_zlim(-1, 4)
		    # plt.title(str(num+1)+"/"+str(no_of_steps_to_animate))
		    ax.set_xlabel(field_name_label[0]) # ax.set_xlabel('P')
		    ax.set_ylabel(field_name_label[1]) # ax.set_ylabel('a1')
		    ax.set_zlabel(field_name_label[2]) # ax.set_zlabel('a2')
		    ax.view_init(elev=view_angle_elevation, azim=view_angle_azimuth)
		    ax.set_title((str(num+1)+"/"+str(no_of_steps_to_animate)))

		    xint = range(int(xlim[0]), int(np.ceil(xlim[1])+1))
		    ax.set_xticks(xint)
		    yint = range(int(ylim[0]), int(np.ceil(ylim[1])+1))
		    ax.set_yticks(yint)
		    zint = range(int(zlim[0]), int(np.ceil(zlim[1])+1))
		    ax.set_zticks(zint)
		    return Q,

		anim = animation.FuncAnimation(fig, update_quiver, fargs=(Q, X, Y, W, ax),
		                               interval=time_between_steps_in_ms, # ms 
		                               blit=False,
		                               repeat=False,
		                               frames= no_of_steps_to_animate, # frames=len(Tset)
		                               )
		
		fig.tight_layout()
		if show_real_time:
			plt.show()
		anim.save(gif_filename, dpi=80, writer='imagemagick')
		return 