import numpy as np

class Rotation3D:
    def __init__(self):
        # convention used: r vector pointing to a point
        # 1. theta : angle between r and z axis
        # 2. phi : angle between r projected to xy plane and x axis
        self.x_set = []
        self.y_set = []
        self.z_set = []
        self.xrot_set = []
        self.yrot_set = []
        self.zrot_set = []

        self.x0 = 0
        self.y0 = 0
        self.z0 = 0
        return

    def rotate_all_towards_z_axis(self):
        # fill in x, y and z sets first
        self.get_centre_coord()
        r0, phi0, theta0 = self.get_spherical_coords(self.x0, self.y0, self.z0)
        if self.check_xyz_sets() is False:
            print('rotate_all_towards_z_axis(). size mismatch for x, y, z sets.')
            return
        for i in range(len(self.x_set)):
            st = np.matrix([[self.x_set[i]], [self.y_set[i]], [self.z_set[i]]])
            # _,  phi, theta = self.get_spherical_coords(st.item(0), st.item(1), st.item(2))
            st_rot = self.rotate_towards_z_axis(st, phi0, theta0) # rotated
            self.xrot_set.append(st_rot.item(0))
            self.yrot_set.append(st_rot.item(1))
            self.zrot_set.append(st_rot.item(2))
        return

    def check_xyz_sets(self):
        out = True
        if len(self.x_set)!= len(self.y_set) or len(self.x_set)!= len(self.z_set) or len(self.y_set)!= len(self.z_set):
            out = False
        return out

    def get_centre_coord(self):
        # fill in x, y and z sets first
        self.x0 = np.mean(self.x_set)
        self.y0 = np.mean(self.y_set)
        self.z0 = np.mean(self.z_set)
        print("center coords  =  ", [self.x0, self.y0, self.z0])
        return

    def rotate_towards_z_axis(self, state, current_phi, current_theta):
        temp = self.rotate_wrt_z(state, -current_phi)
        out = self.rotate_wrt_y(temp, -current_theta)
        return out

    def get_spherical_coords(self, x, y, z, smallxy=True, precisionxy=1e-15):
        r = (x**2 + y**2 + z**2) ** 0.5
        if smallxy:
            if x <precisionxy and y<precisionxy:
                phi = 0.0
            else:
                phi = np.arctan(y / x)
        else:
            phi = np.arctan(y / x)
        theta = np.arccos(z/r)
        return r, phi, theta

    def rotate_wrt_x(self, state, angle):
        # state: 3 by 1 matrix
        Ay = [[1,0, 0],
              [0, np.cos(angle), np.sin(angle)],
              [0, -np.sin(angle), np.cos(angle)]]
        out = np.dot(Ay, state)
        return out

    def rotate_wrt_y(self, state, angle):
        # state: 3 by 1 matrix
        Ay = [[np.cos(angle), 0, np.sin(angle)],
              [0, 1, 0],
              [-np.sin(angle), 0, np.cos(angle)]]
        out = np.dot(Ay, state)
        return out

    def rotate_wrt_z(self, state, angle):
        # state: 3 by 1 matrix
        Az = np.matrix([[np.cos(angle), -np.sin(angle), 0],
                        [np.sin(angle), np.cos(angle), 0],
                        [0,0,1]])
        out = np.dot(Az, state)
        return out